% --- Source file: crave.Rd ---
\name{crave}
\alias{crave}
\title{crave}
\description{
  A function for detecting associations between rare variant SNPs and phenotypes
}
\usage{
crave(geno_file, pheno_file, out_gene, out_snp, op=NULL)
}
\arguments{
  \item{geno_file}{The file containing the genotype data (either a TPED or VCF file). 
                   This file may be compressed with gzip.}
  \item{pheno_file}{The file containing the response and covariate data. If \code{geno_file} is a VCF file, then
                    \code{pheno_file} must contain a column of subject ids.}
  \item{out_gene}{Output file for the gene results.}
  \item{out_snp}{Output file for the SNP results.}
  \item{op}{List of options. See \code{details}.}
}
\value{
  The returned value is NULL. All final results are stored in \code{out_gene} and \code{out_snp}.
 }
\details{  

\bold{Genotype data:} \cr
The first 4 columns of a TPED file should be the chromosome, SNP id, distance, and location.
      The distance in column 3 is not used, and the chromosome should be labeled as:
      1-22, 23 or X, 24 or Y, 25 or XY, 26 or MT.
      Starting in column 5, are the genotypes for each subject.
      The order of these genotypes must match the order of the subjects in the phenotype file.
      An example row of a TPED file would be:\cr
      1 rs3132489 0 2345643 A A A G A A G G 0 0 A A A G \cr
      In the row above, 0 denotes a missing allele. Set the option \code{allele_miss} to specify a different missing value. 
   
      A VCF file must be grouped by chromosome and SNPs sorted by increasing location within each chromosome.
      The first 9 columns of a VCF file must be:\cr
      #CHROM, POS, ID, REF, ALT, QUAL, FILTER, INFO, FORMAT \cr
      Starting in column 10 are the subject ids. No order is assumed for these ids,
      however some of the ids must match the ids in the phenotype file. Subjects with un-matched ids will be removed
      from the anlaysis.

\bold{Phenotype data:} \cr
Any subject with a missing value for the response or a covariate will be removed from the analysis.
The response can either be continuous or binary. If the response is binary and not coded as 0 for controls and 1 for cases, 
then use the options \code{control_value} and \code{case_value}.

  \bold{Options list:} \cr
 Below are the names for the options list \code{op}. All names have default values
         if they are not specified.
\itemize{
  \item{\code{adapt_maxperm}}{ Maximum number of adaptive permutations. Set to 0 for non-adaptive permutations.
                          The stopping criteria (see \code{adapt_tol}) is checked after every \code{nperm} permutations.      
                                  The default value is 1000000.} 

  \item{\code{adapt_tol}}{ Stopping tolerance for adaptive permutations. \cr
            The algorithm stops if 3.92*sqrt(pval*(1.0-pval)/N)/pval < \code{adapt_tol}, 
                       where N is the total number of permutations that have been run so far.
                                  The default is 0.1.}

  \item{\code{addcols}}{ 0-2. Specifying 1 or 2 will add columns to the \code{out_gene} file. The additional columns
           are the number of permutations and number of times that the permutation test statistic was
           greater than the observed.
                       The default is 0. }

  \item{\code{allele_miss}}{ Missing value for the alleles in a TPED file. The default value is 0. }

  \item{\code{bm_n}}{ Number of points to consider in the Bounded Minimum Test.
                                  The default is 10.}

  \item{\code{case_value}}{ Integer value of the response column denoting the cases. 
                            The default is 1.}
  \item{\code{chr_alias}}{ Tab delimited text file with the first column as the chromosome number (1-26) and the second column 
                           the chromosome string. This option is only used with VCF genotype files when the chromosome fields 
                           are not labeled as integers.}
  \item{\code{control_value}}{ Integer value of the response column denoting the controls. 
                            The default is 0.}

  \item{\code{covars}}{ Vector of column names or column numbers in \code{pheno_file} to include
              as covariates. The default is that no covariates will be used.}

  \item{\code{das_alpha}}{ A number between 0 and 1 for the alpha parameter in the Data Adaptive Sum Test.
                                  The default is 0.05.}

  \item{\code{exclude_chr}}{ Vector of chromosome numbers to exclude in anlaysis. Example: 1-6,13,22.
                                  The default value is NULL. }

  \item{\code{exclude_gene}}{ File containing 1 column of the genes to exclude in the analysis.}

  \item{\code{exclude_snp}}{ File containing 1 column of the SNPs to exclude in the analysis.}

  \item{\code{exclude_sub}}{ File containing 1 column of the subject ids to exclude in the analysis.}

  \item{\code{gene_file}}{ Folder containing the gene information files. 
            To use hg16, hg17, or hg18, set gene_file=folder containing the gene information files.
            By default, hg19 is used and is installed with CRaVe. 
            The files are named chr_xxx.txt.gz.
            Each file contains the columns: \cr
            Column 1: gene name \cr
            Column 2: gene start position \cr
            Column 3: gene end position \cr
            Column 4: codon start position \cr
            Column 5: codon end position \cr
            Column 6: number of exon start and end positions \cr
            Column 7: comma seperated list of exon starting positions \cr
            Column 8: comma seperated list of exon ending positions \cr
            The files must be tab-delimited, sorted by ascending gene start position.
            }
  \item{\code{gene_kb}}{ Number of kilobases upstream and downstream to expand the gene regions in the gene tables. The default is 10.}

  \item{\code{geno_gz}}{ 0 or 1 for an uncompressed/compressed genotype file with gzip. The default is determined from the file extension.}
  \item{\code{geno_sep}}{ File delimiter for the genotype file.}

  \item{\code{id}}{ Column name or number of the id variable in \code{pheno_file}.
                    The default value is 1. }

  \item{\code{include_chr}}{ Vector of chromosome numbers to include in anlaysis. Example: 1-6,13,22.
                                  The default value is 1:26 }
  
  \item{\code{include_gene}}{ File containing 1 column of the genes to include in the analysis.}

  \item{\code{include_snp}}{ File containing 1 column of the SNPs to include in the analysis.}
  
  \item{\code{include_sub}}{ File containing 1 column of the subject ids to include in the analysis.}
  
  \item{\code{max_maf}}{ The maximum MAF for a SNP to be included in the analysis, so that any SNP with MAF > \code{max_maf}
                          will be excluded.
                               The default is 0.5. }
  \item{\code{max_miss_rate}}{ Maximum missing rate for any SNP to be included in the analysis.
                                  The default value is 0.8. }
  
  \item{\code{min_maf}}{ The minimum MAF for a SNP to be included in the analysis, so that any SNP with MAF < \code{min_maf}
                          will be excluded.
                               The default is 0. }

  \item{\code{notGroupedByChr}}{ Set to TRUE if \code{geno_file} is a TPED file and this file
   is not grouped by chromosome. Note that a VCF file should always be grouped by chromosome.
   The default value is FALSE. }
  
  \item{\code{no_gene_file}}{ Set to TRUE to not use the default gene table file for the gene names.
                                  Instead use the \code{reflink.geneName} value in the VCF file.
                                  This option is only valid if \code{geno_file} is a VCF file and the \code{weight=} option
                                  is not specified.
                                  The default is FALSE. }

  \item{\code{nperm}}{ The number of permutations to run for each iteration of the adaptive permutation algorithm.
                                   The default is 500.}

  \item{\code{pheno_header}}{ 0 or 1 if \code{pheno_file} contains column names. }

  \item{\code{pheno_miss}}{ Missing value for the \code{pheno_file}.
                                  The default value is '.'}
  
  \item{\code{pheno_sep}}{ File delimiter for the phenotype file. }
  
  \item{\code{print}}{ The larger the value, the more information will be printed to the \code{log_file}.
                       The default is 0 unless \code{log_file} is specified, in which case the default is 2. }

  \item{\code{reflink.geneName}}{ Name of the variable in the INFO field of the VCF file that gives the gene name.
                                  The default value is "reflink.geneName"}

  \item{\code{response}}{ Column name or number of the response variable in \code{pheno_file}.
                    The default value is 2. }

  \item{\code{rover_t}}{ Positive number t to define the weights (w = 1-exp(-t*delta*delta)) in the Rover Test.
                                  The default is 0.2}

  
  \item{\code{seed}}{ Random seed. The default will be computed from the time. }

  \item{\code{sortedByLoc}}{ 0 or 1. Set to 0 if the genotype data is not sorted by SNP location within each chromosome.
                                    The default is 1. }

  \item{\code{store_perms}}{ 0 or 1. Set to 1 to store the permutations. Due to memory issues, this option should only
                                    be used for small sample sizes and number of permutations.
                                    The default is 0. }

  \item{\code{tests}}{ Character vector of tests to use. The possible tests are:\cr
      ALL  - To compute all non user defined tests \cr
      BM   - Bounded Minimum Test \cr
      CMC  - Combined Multivariate and Collapsing Test \cr
      DAS  - Data Adaptive Sum Test \cr
      FI   - Fisher's Test \cr
      HOI  - Hotelling's T^2 Test assuming independence \cr
      HOT  - Hotelling's T^2 Test \cr
      MDF  - Similarity Regression Test \cr
      MDP  - Similarity Regression Test using positive deltas \cr
      RO   - ROVER (Reduction of Variables Explained by Random noise) Test \cr    
      RV   - ROVER-V Rover Test that includes the variances in the weights \cr 
      STO  - Stouffer's Z-score Test \cr
      STP  - Stouffer's Positive Z-score Test \cr
      SUM  - Sum Test \cr
      TH   - Threshold Test \cr
      SKAT - Alias for MDF [Under standard conditions, SKAT [PMID: 21737059] is equivalent to MDF] \cr
      CAL  - Alias for C-alpha [Under standard conditions, the C-alpha statistic [PMID: 21408211] is equivalent to MDF] \cr

                                  Use the strings "c1", "c2", etc to specify user-defined C functions (see below).
                                  The default is to use the MDF test.}
  
  \item{\code{threshold_alpha}}{ A number between 0 and 1 for the alpha parameter in the Threshold Test.
                                  The default is 0.05.}

  \item{\code{update_cor}}{ Value such that the correlation matrix will be recomputed if the relative difference in the observed and
  permuted Hotelling test statistics is less than \code{update_cor}. A large value for \code{update_cor} will cause the correlation matrix and it's inverse to be recomputed in each permutation. A value less than or equal to 0 for \code{update_cor} will allow for 
quicker computation times. This option is only valid for case-control data. 
  The default is 0.3. }

  \item{\code{user_func}}{ Character vector of user-defined R function names to be used as test statistics. A maximum of
                           5 user-defined tests are allowed. See below for user-defined C functions.}

  \item{\code{vs}}{ The value of \code{vs}, where \code{vs} abbreviates variance stabilization, is the value added to the estimated variance of each SNP.                   The default is 0.005}

  \item{\code{weight}}{ "e", "c", or "n". weight=e assigns a weight of 1 to all exomic SNPs and 0 to other SNPs.
                              weight=c assigns a weight of 1 to SNPs between the start and stop codon and 0 to other SNPs.
                              weight=n assigns a weight of 1 to non-synonomous exomic SNPs and 0 to other SNPs.
                              The default is that a weight of 1 is assigned to all SNPs.
                              This option cannot be used with the \code{weight_file} option.}

  \item{\code{weight_file}}{ File containing SNP weights. This file may have 1 or 2 columns only. 
            For a 1 column file the column of weights must match the order of the SNPs in the genotype file. For a 2 column file, the first column is the SNP               id and the second column the weight.}

  \item{\code{weight_header}}{ Set to 1 if row 1 of the weight file contains column names. The default value is determined from the file.}

  \item{\code{weight_sep}}{ One character value for the weight file delimiter.}


  
  %\item{\code{log_file}}{ The file where messages will be stored. If \code{print} = 0, then \code{log_file}
  %                        will not be created. If \code{print} > 0 and \code{log_file} is NULL, then a default
  %                        log file will be created with name paste(\code{out_file}, ".LOG", sep=""). }
  
  
  
  


} % END: itemize

\bold{User-defined tests:} \cr
Any user-defined R function to compute a test statistic must return a single number and have 4 input arguments. Argument 1
is the vector of scores for the SNPs in a gene, argument 2 is the gene covariance matrix, argument 3 is the gene
correlation matrix, and argument 4 is the inverse correlation matrix for the gene. The results for these tests are in 
\code{out_file} with prefix "USER_R1", "USER_R2", etc, where the order corresponds to the order of the functions listed
in the option \code{user_func}.
A user-defined C function to compute a test statistic must be defined in the file \code{user_defined.c} in the "src" folder.
Up to 5 user-defined C functions can be written, however the names of these functions must be "c1", "c2", "c3", "c4", "c5", and
the order/number of input arguments cannot be modified. See the function "c1" in \code{user_defined.c}.


} % END: details

%\references{ }
%\author{}
%\seealso{ }
\examples{
 
 geno_file <- system.file("sampleData", "sim_tped.txt", package="CRAVE")
 pheno_file <- system.file("sampleData", "sim_pheno.txt", package="CRAVE")
 out_gene <- "c:/temp/out_gene.txt"
 out_snp <- "c:/temp/out_snp.txt"

 # Run with the defaults
 #crave(geno_file, pheno_file, out_gene, out_snp, op=list(tests=c("RO", "SUM")))

 # User defined R function
 fnc1 <- function(delta, cov, cor, invcor) {
   return(sum(delta*delta))
 }

 # Include the user-defined test
 #crave(geno_file, pheno_file, out_gene, out_snp, op=list(user_func="fnc1", print=3))

}

\keyword{model} 

