/***************************************************************
                                                             
                    CRAVE                  
                                                              
 This file is distributed under the GNU General Public        
 License, Version 2. See the file COPYING.txt for more    
 details                                                      
                                                              
****************************************************************/

#include "./crave.h"

static char *main_rfunc;
static char *user_rfunc1;
static char *user_rfunc2;
static char *user_rfunc3;
static char *user_rfunc4;
static char *user_rfunc5;
static char *print_rfunc;


/* Main function for user-defined R functions.
   User-defined C function: u1(delta, cov, cor, invcor, nsnp)
 */
double user_r_main(info, GENE, which)
INFO *info;
GENE_STR *GENE;
int which;
{
  char *modes[5];
  char *arguments[5];
  double *result, *p1, *p2, *p3, *p4, *p5, *p6, *p7, *p8, **cov, **cor, **inv;
  long lengths[5];
  int i, j, nsnp=GENE->n_snps, nsnp2;

  /* Copy to vectors */
  cov = info->geno_cov;
  cor = info->geno_cor;
  inv = info->geno_invcor;
  p3 = info->user_cov;
  p5 = info->user_cor;
  p7 = info->user_invcor;
  for (i=0, p1=info->delta, p2=info->user_delta; i<nsnp; i++, p1++, p2++) {
    *p2 = *p1;
    for (j=0, p4=cov[i], p6=cor[i], p8=inv[i]; j<nsnp; j++, p4++, p6++, p8++) {
      *p3++ = *p4;
      *p5++ = *p6;
      *p7++ = *p8;
    }
  } 
  nsnp2 = nsnp*nsnp;

  lengths[0]   = (long) nsnp;
  arguments[0] = (char *) info->user_delta;
  modes[0]     = "double";
  lengths[1]   = (long) nsnp2;
  arguments[1] = (char *) info->user_cov;
  modes[1]     = "double";
  lengths[2]   = (long) nsnp2;
  arguments[2] = (char *) info->user_cor;
  modes[2]     = "double";
  lengths[3]   = (long) nsnp2;
  arguments[3] = (char *) info->user_invcor;
  modes[3]     = "double";
  lengths[4]   = (long) 5;
  modes[4]     = "character";
  switch (which) {
    case TEST_R1:
      arguments[4] = user_rfunc1;
      break;
    case TEST_R2:
      arguments[4] = user_rfunc2;
      break;
    case TEST_R3:
      arguments[4] = user_rfunc3;
      break;
    case TEST_R4:
      arguments[4] = user_rfunc4;
      break;
    case TEST_R5:
      arguments[4] = user_rfunc5;
      break;
  }
  
  call_R(main_rfunc, (long) 5, arguments, modes, lengths, (char *) 0, (long) 1, arguments);
  result = (double *) arguments[0];

  return(*result); 

} /* END: user_r_main */ 

int print_r_main(str0, str, pdbl, pint, pchar)
char *str0, *str;
double *pdbl;
int *pint;
char *pchar;
{
  char *modes[5];
  char *arguments[5], pc[2], *pstr;
  long lengths[5];
  int which, ival=0;
  double dval=0.0;

  if (str) {
    which = 1;
    pstr = str;
  } else if (pdbl) {
    which = 2;
    dval = *pdbl;
    pstr = "";
  } else if (pint) {
    which = 3;
    ival = *pint;
    pstr = "";
  } else if (pchar) {
    which = 1;
    pc[0] = *pchar;
    pc[1] = END_OF_STR;
    pstr = pc;
  } else {
    which = 0;
    pstr = "";
  }

  arguments[0] = (char *) str0;
  modes[0]     = "character";
  lengths[0]   = (long) 1;
  arguments[1] = (char *) &which;
  modes[1]     = "integer";
  lengths[1]   = (long) 1;
  arguments[2] = (char *) pstr;
  modes[2]     = "character";
  lengths[2]   = (long) 3;
  arguments[3] = (char *) &dval;
  modes[3]     = "double";
  lengths[3]   = (long) 1;
  arguments[4] = (char *) &ival;
  modes[4]     = "integer";
  lengths[4]   = (long) 1;

  call_R(print_rfunc, (long) 5, arguments, modes, lengths, (char *) 0, (long) 1, arguments);

  return(0); 

} /* END: print_r_main */ 

int mymain(argc, argv)
int argc;
char *argv[];
{
  INFO info;
  int print;
  double t11=0.0;

  #if RFLAG==0
  clock();
  #endif

  /* Initialize the info structure and parse the user options */
  FPT_LOG_FILE = NULL;
  init(&info, argv, argc);
  print = info.print;

  /* Check for a log file and open it */
  if (info.log_file != NULL) FPT_LOG_FILE = open_file(info.log_file, "w");

  if (print) {
    printf_s("-----\n", NULL);
    printf_s("Begin\n", NULL);
    printf_s("-----\n", NULL);
  }

  /* Open the output file */
  open_out(&info);
  if (print > 1) print_info(&info, 1);

  /* Open the genotype data */
  my_fopen(info.geno_file, &info.gfpt);

  /* Get the phenotype data */
  if (print) {
    printf_s("-----------------------\n", NULL);
    printf_s("Load the phenotype data\n", NULL);
    printf_s("-----------------------\n", NULL);
  }
  getPheno2(&info);
  if (print > 1) print_info(&info, 2);

  /* Get the columns in the genotype file and check them */
  if (print) {
    printf_s("-----------------------------------------------\n", NULL);
    printf_s("Load and check the columns in the genotype file\n", NULL);
    printf_s("-----------------------------------------------\n", NULL);
  }
  getCols(&info);
  if (print > 1) print_info(&info, 3);

  update_info(&info);
  if (print > 1) print_info(&info, 4);

  /* Process the snps */
  if (print) {
    printf_s("-------------------------------\n", NULL);
    printf_s("Begin reading the genotype data\n", NULL);
    printf_s("-------------------------------\n", NULL);
  }
  read_geno(&info);

  /* Output genes in last chromosome */
  chrPermsOutput(&info);
  
  my_fclose(&info.gfpt);
  if (print) {
    printf_s("-------------------\n", NULL);
    printf_s("Free info structure\n", NULL);
    printf_s("-------------------\n", NULL);
  }
  free_info(&info);

  /* Total time */
  #if RFLAG==0
  t11 = (double) clock()/CLOCKS_PER_SEC;
  if (print) {
    if (t11 < 60.0) {
      printf_d("\nTotal time is %.0f seconds\n", t11);
    } else if (t11 < 3600.0) {
      printf_d("\nTotal time is %.1f minutes\n", t11/60.0);
    } else {
      printf_d("\nTotal time is %.2f hours\n", t11/3600.0);
    }
  }
  #endif

  if (FPT_LOG_FILE) fclose(FPT_LOG_FILE);

  return(0);
}

/* Function called from R */
void crave(argc, argv, user, main_f, print_f, user_f, n_f, ret)
int *argc;
char **argv;
int *user;      /* 0 or 1 for a user-defined R function */
char **main_f;  /* address of crave_user_func */  
char **print_f; /* address of crave_print */  
char **user_f;  /* Character vector of user-defined R function names */
int *n_f;       /* Length of user_f */
int *ret;
{
  int nf;

  *ret = 1;

  print_rfunc  = print_f[0];
  if (*user == 1) {
    main_rfunc  = main_f[0];
    user_rfunc1 = user_f[0];
    nf = *n_f;
    if (nf > 1) user_rfunc2 = user_f[1];
    if (nf > 2) user_rfunc3 = user_f[2];
    if (nf > 3) user_rfunc4 = user_f[3];
    if (nf > 4) user_rfunc5 = user_f[4];
  }
  *ret = mymain(*argc, argv);
  return;

}

