/***************************************************************
                                                             
                    CRAVE                  
                                                              
 This file is distributed under the GNU General Public        
 License, Version 2. See the file COPYING.txt for more    
 details                                                      
                                                              
****************************************************************/

#include "./crave.h"

static int myinv(cov, n, inv)
double cov[2][2], inv[2][2];
int n;
{
  double a, b, c, d, cc;

  a  = cov[0][0];
  b  = cov[0][1];
  c  = cov[1][0];
  d  = cov[1][1];
  cc = a*d - b*c;
  if (fabs(cc) < ALMOST_ZERO) return(ERROR_SINGULAR_MATRIX);
  cc = 1.0/cc;
  inv[0][0] = d*cc;
  inv[0][1] = -b*cc;
  inv[1][0] = -c*cc;
  inv[1][1] = a*cc;

  return(0);

} /* END: myinv */ 

/* Function to compute beta = inv(X`X)X`y for simple case: intercept and snp */
static int beta_norm(y, snp, nr, beta0, beta1, xpx)
double *y, *beta0, *beta1;
int nr;
double xpx[2][2];
genoType *snp;
{
  double inv[2][2], sum1=0.0, sum2=0.0, *pd, temp;
  int ret, i;
  genoType *pg, gg;

  /* Compute the inverse */
  ret = myinv(xpx, 2, inv);

  /* Compute X`y */
  for (i=0, pd=y, pg=snp; i<nr; i++, pd++, pg++) {
    gg = *pg;
    if (gg != MISS_GENO) {
      temp = *pd;
      sum1 += temp;
      sum2 += temp*gg;
    } 
  }

  /* Compute beta */
  *beta0 = inv[0][0]*sum1 + inv[0][1]*sum2;
  *beta1 = inv[1][0]*sum1 + inv[1][1]*sum2;

  return(ret);

} /* END: beta_normal */


/* Function to compute beta for logistic reg onj intercept and 0-1-2 snp */
static int beta_cc(controls, snp, nr, xpx, beta0, beta1, z)
binary *controls;  
double *beta0, *beta1, xpx[2][2], *z;
int nr;
genoType *snp;
{
  double mat[2][2], inv[2][2], beta00, beta01, b0, b1, xb0, xb1, xb2, sum0, sum1, sum2, pvec0, pvec1, pvec2;
  double vec0, vec1, vec2, temp1, temp2, temp3, temp4, vec20, vec21;
  int ii, ret=0, n00, n01, n02, n10, n11, n12, n00_n10, n01_n11, n02_n12, N, counts[2][3];  
  genoType gg, *pg;
  binary *pc;

  /* Initial estimates [(X`X)^-1]X`Z, where Z = 2Y-1 */
  ret = beta_norm(z, snp, nr, &beta00, &beta01, xpx);
  if (ret) return(ret);
   
  /* Get the counts */
  counts[0][0] = counts[0][1] = counts[0][2] = counts[1][0] = counts[1][1] = counts[1][2] = 0;
  for (ii=0, pc=controls, pg=snp; ii<nr; ii++, pc++, pg++) {
    gg = *pg;
    if (gg != MISS_GENO) counts[1-*pc][gg]++;
  }
  n00     = counts[0][0];
  n01     = counts[0][1];
  n02     = counts[0][2];
  n10     = counts[1][0];
  n11     = counts[1][1];
  n12     = counts[1][2];
  n00_n10 = n00 + n10;
  n01_n11 = n01 + n11;
  n02_n12 = n02 + n12;
  N       = n00_n10 + n01_n11 + n02_n12;

  /* Loop until convergence */
  for (ii=0; ii<LOGREG_MAXIT; ii++) {

    /* Compute pi */
    xb0   = beta00;
    xb1   = beta00 + beta01;
    xb2   = xb1 + beta01;
    sum0  = 1.0/(1.0 + exp(-xb0));
    sum1  = 1.0/(1.0 + exp(-xb1));
    sum2  = 1.0/(1.0 + exp(-xb2));
    pvec0 = sum0;
    pvec1 = sum1;
    pvec2 = sum2;
    sum0  = sum0*(1.0 - sum0);
    sum1  = sum1*(1.0 - sum1);
    sum2  = sum2*(1.0 - sum2);
    if (sum0 < LOGREG_ZERO) {
      sum0 = LOGREG_ZERO;
    } else if (sum0 > LOGREG_ONE) {
      sum0 = LOGREG_ONE;
    }
    if (sum1 < LOGREG_ZERO) {
      sum1 = LOGREG_ZERO;
    } else if (sum1 > LOGREG_ONE) {
      sum1 = LOGREG_ONE;
    }
    if (sum2 < LOGREG_ZERO) {
      sum2 = LOGREG_ZERO;
    } else if (sum2 > LOGREG_ONE) {
      sum2 = LOGREG_ONE;
    }
    vec0 = sum0;
    vec1 = sum1;
    vec2 = sum2;

    temp1 = n01_n11*vec1;
    temp2 = n02_n12*vec2;
    mat[0][0] = n00_n10*vec0 + temp1 + temp2 + LOGREG_FUZZ;
    mat[0][1] = mat[1][0] = temp1 + temp2 + temp2; 
    mat[1][1] = temp1 + 4*temp2 + LOGREG_FUZZ;
    temp1 = n01*pvec1;
    temp2 = n02*pvec2;
    temp3 = n11*(1.0 - pvec1);
    temp4 = n12*(1.0 - pvec2);
    vec20 = -(n00*pvec0 + temp1 + temp2) + n10*(1.0 - pvec0) + temp3 + temp4;
    vec21 = -(temp1 + temp2 + temp2) + temp3 + temp4 + temp4;

    /* Invert mat */
    ret = myinv(mat, 2, inv);
    if (ret) break;

    /* Update beta, beta0 and check convergence */
    temp1  = inv[0][0]*vec20 + inv[0][1]*vec21;
    temp2  = inv[1][0]*vec20 + inv[1][1]*vec21;
    b0     = temp1 + beta00;
    b1     = temp2 + beta01;
    temp3  = fabs(temp1) + fabs(temp2);
    if (temp3 < LOGREG_EPS) break;
    beta00 = b0;
    beta01 = b1;
  }   
  if (ii >= LOGREG_MAXIT) ret = ERROR_MAXITER;

  *beta0 = b0;
  *beta1 = b1;

  return(ret);

} /* END: beta_cc */

/* Function to compute all the effect sizes for the snps in a gene */
void effect_size(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, j, nsnp=GENE->n_snps, nsub=info->nsubs;
  double *response=NULL, beta0, beta1, xpx[2][2], *pd, *pd2;
  SNP_STR *SNP=GENE->first;
  yType *y=info->CC;
  binary CC_flag=info->CC_flag, *controls=info->controls;
  
  if (CC_flag) {
    response = (double *) malloc(nsub*sizeof(double));
    CHECK_MEM(response);
    for (i=0, pd=y, pd2=response; i<nsub; i++, pd++, pd2++) *pd2 = 2.0* *pd - 1.0;
  }

  for (i=0; i<nsnp; i++) {
    xpx[0][0] = nsub - SNP->nmiss;
    xpx[0][1] = xpx[1][0] = SNP->n1 + 2*SNP->n2;
    xpx[1][1] = SNP->n1 + 4*SNP->n2; 
  
    if (CC_flag) {
      j = beta_cc(controls, SNP->G, nsub, xpx, &beta0, &beta1, response);
      if (!j) beta1 = exp(beta1);
    } else {
      j = beta_norm(y, SNP->G, nsub, &beta0, &beta1, xpx);
      if (!j) beta1 = exp(beta1);
    }

    if (!j) {
      SNP->effect_size = beta1;
    } else {
      SNP->effect_size = ERROR_EFFECT_SIZE;
    }
    SNP = SNP->next;
  }

  if (response) free(response);

} /* END: effect_size */




