/***************************************************************
                                                             
                    CRAVE                  
                                                              
 This file is distributed under the GNU General Public        
 License, Version 2. See the file COPYING.txt for more    
 details                                                      
                                                              
****************************************************************/

#include "./crave.h"

/* Function to print help about a specific option */
void print_option(which)
int which;
{
  switch (which) {
    case 0:
      printf("pheno_file= This file must be an uncompressed text file containing the outcome and possibly covariates and a column of ids.\n");
      printf("            The column of ids is required if geno_file is a VCF file.\n");
      printf("            Use the options response=, id=, and covars= to specify variables.\n");
      printf("            If the file contains covariates, then they will be included in the analysis by default.\n");
      printf("            Subjects with missing values for the outcome or any included covariate will be excluded from the analysis.\n");  
      printf("            Use the pheno_miss= option to specify the missing value if not a '.'.\n");
      printf("            For case-control studies, the response must be integer valued.\n");
      printf("            The subject ids can be anything if the genotype file is a TPED file, however\n");
      printf("            the number of subjects in the phenotype file must equal the number in the TPED file and the order must be the same.\n");
      printf("            If the genotype file is a VCF file, then the number of subjects and order do not matter.\n");
      printf("            By default, this file is assumed to be tab delimited. For different delimiters, set pheno_sep=\n");
      break; 
    case 1:
      printf("geno_file= File containing the genotype data.\n");
      printf("           This file can either be in VCF format or in a TPED format\n");
      printf("           A VCF file must be grouped by chromosome and sorted by increasing location within each chromosome.\n");
      printf("           The first 9 columns of a VCF file must be:\n");
      printf("           #CHROM, POS, ID, REF, ALT, QUAL, FILTER, INFO, FORMAT\n");
      printf("           Starting in column 10 are the subject ids. No order is assumed for these ids,\n");
      printf("           however some of the ids must match the ids in the phenotype file.\n");
      /*
      printf("           A VCF file may contain the names reflink.geneName, refseq.changesAA, and refseq.functionalClass in the INFO column.\n");
      printf("           If so, then the gene_file option does not need to be used, and each SNP will be assigned a weight\n");
      printf("           as a function of refseq.changesAA and refseq.functionalClass.\n");
      printf("           The variables reflink.geneName, refseq.changesAA, and refseq.functionalClass can be obtained from the GenomicAnnotator GATK tool.\n");
      */
      printf("           The first 4 columns of a TPED file should be the chromosome, SNP id, distance, and location.\n");
      printf("           The distance in column 3 is not used, and the chromosome should be labeled as:\n");
      printf("           1-22, 23 or X, 24 or Y, 25 or XY, 26 or MT\n");
      printf("           Starting in column 5, are the genotypes for each subject.\n");
      printf("           The order of these genotypes must match the order of the subjects in the phenotype file\n");
      printf("           An example row of a TPED file would be:\n");
      printf("           1 rs3132489 0 2345643 A A A G A A G G 0 0 A A A G\n");
      printf("           In the row above, 0 denotes a missing allele. Set the option allele_miss= to specify a different missing value\n");
      break;
    case 2:
      printf("out_gene= The output file will be a tab delimited file containing the final permutation p-values for each gene.\n");
      printf("          The columns in this file will be for the gene, chromosome, gene start location, gene end location, number of SNPs, and p-values for each test.\n");
      printf("          The number of SNPs refers the number of SNPs in the gene that were used in the analysis.\n");
      break;
    case 3:
      printf("max_maf= This value determines which SNPs will be included in the analysis.\n");
      printf("         Any SNP with MAF > max_maf will be discarded. The MAF for each SNP is determined from using all the non-missing alleles.\n");
      printf("         The default value for max_maf is 0.5.\n");
    case 4:
      printf("nperm= The number of permutations to be run for each iteration of the adaptive permutation algorithm or the total number of permutations.\n");
      printf("       The SNPs are permuted jointly.\n");
      printf("       The default value is 500.\n");
    case 5:
      printf("  seed=<integer>                  Initial seed for the random permutations\n");
      printf("                                  The default is the returned value from the time() function\n");
      break;
    case 6: 
      print_help(0);
      break;
    case 7:
      printf("pheno_sep= The delimiter in the phenotype file. By default, pheno_file is assumed to be tab delimited.\n");
      printf("           This delimiter must be a single character.\n");
      printf("           Other possible delimiters are a space ' ' and a comma ','.\n");
    case 8:
      printf("  print=<integer>                 The larger the integer, the more information will be printed\n");
      printf("                                  The default is 0 (no printing)\n");
      break;
    case 9:
      printf("  reflink.geneName=<string>       Name of the variable in the INFO field of the VCF file that gives the gene name\n");
      printf("                                  The default value is reflink.geneName\n");
      break;
    case 10:
      break;
    case 11:
      break;
    case 12:
      break;
    case 13:
      break;
    case 14:
      break;
    case 15:
      break;
    case 16:
      break;
    case 17:
      printf("  refseq.changesAA=<string>       Name of the variable in the INFO field of the VCF file that gives the value of refseq.changesAA\n");
      printf("                                  The default value is refseq.changesAA\n");
      break;
    case 18:
      printf("  refseq.functionalClass=<string> Name of the variable in the INFO field of the VCF file that gives the value of refseq.functionalClass\n");
      printf("                                  The default value is refseq.functionalClass\n");
      break;
    case 19:
      printf("gene_file=  This option is required for TPED genotype files and can also be used for VCF genotype files.\n");
      /*
      printf("            When this option is used, all SNPs get assigned a weight of 1.\n");
      */
      break;
    case 20:
      printf("  --version                       Display the version number and exit\n");
      break;
    case 21:
      printf("--notGroupedByChr   Use this option if geno_file is a TPED file which is not grouped by chromosome.\n");
      printf("                    This option will require more memory to be used, since the entire genotype file must \n");
      printf("                    first be read in and then processed.\n");
      printf("                    This option cannot be used with VCF files, since VCF files must be grouped by chromosome.\n");
      break;
    case 22:
      break;
    case 23:
      break;
    case 24:
      break;
    case 25:
      break;
    case 26:
      printf("tests=    The possible tests are:\n");
      printf("          ALL  - To compute all non user defined tests\n");      
      printf("          BM   - Bounded Minimum Test\n");
      printf("          CMC  - Combined Multivariate and Collapsing Test\n");
      printf("          DAS  - Data Adaptive Sum Test\n");
      printf("          FI   - Fisher's Test\n");
      printf("          HOI  - Hotelling's T^2 Test assuming independence\n");
      printf("          HOT  - Hotelling's T^2 Test\n");
      printf("          MDF  - Multiple Degress of Fredom Test\n");
      printf("          MDP  - Multiple Degrees of Freedom Test using positive deltas\n");
      printf("          RO   - ROVER (Reduction of Variables Explained by Random noise) Test\n");    
      printf("          RV   - ROVER-V Rover Test that includes the variances in the weights\n");   
      printf("          STO  - Stouffer's Z-score Test\n");
      printf("          STP  - Stouffer's Positive Z-score Test\n");
      printf("          SUM  - Sum Test\n");
      printf("          TH   - Threshold Test\n");      
      printf("          SKAT - Same as MDF\n");  
      printf("          CAL  - Same as MDF\n");
      printf("          C1   - User defined test\n");
      printf("          To create a user defined test, edit a function in the file user_defined.c\n");
      printf("          Up to 5 user defined tests can be created.\n");
      printf("          Setting tests=all,c1,c2,c3,c4,c5 will compute all tests.\n");
      break;
    case 27:
      break;
    case 28:
      printf("bm_n=   This integer breaks up the interval [0, 4] into equally spaced points to compute the minimum p-value.\n");
      printf("        If bm_n = 1 or 2, then the points are 2 and 0, 4 respectively.\n"); 
      printf("        For bm_n > 2, the points are i*4/bm_n, i=0, 1,..., bm_n-1.\n");
      break;
    case 29:
      break;
    case 56:
      printf("adapt_maxperm=    Maximum number of permutations for adaptive p-values.\n");
      printf("                  For each iteration of the adaptive permutation algorithm, nperm permutations are run and then the p-value is computed.\n");
      printf("                  The algorithm stops if 3.92*sqrt(pval*(1.0-pval)/N)/pval < adapt_tol,\n");
      printf("                  where N is the total number of permutations that have been run so far.\n");
      break;
    case 57:
      printf("out_snp= The output file will be a tab delimited file containing information for each SNP used in the analysis.\n");
      printf("          The columns in this file will be for the chromosome, gene, SNP position, MAF, number of non missing genotypes, effect size, delta and p-value.\n");
      printf("          The number of SNPs refers the number of SNPs in the gene that were used in the analysis.\n");
      break;
    case 60:
      printf("update_cor= The correlation matrix will be recomputed with each permutation if update_cor is set to a large value.\n");
      printf("            Conversely, if update_cor is set to 0 or a negative value, the correlation matrix will not be recomputed.\n");
      printf("            This option is only valid for case-control data and when the option --useAllSubsForV is not used.\n");
      break;
    case 67:
      printf("chr_alias_file= In VCF files, the chromosome field is not always an integer value (1-26).\n");
      printf("            This option allows the user to define the mapping from the chromosome field to a chromosome number.\n");
      printf("            This option is only valid for VCF genotype files.\n");
      break;
    default:
      break;

  } /* END: switch */

  exit(0);

} /* END: print_option */

/* Function to print the help message */
void print_help(which)
int which;
{
  printf("Usage: crave [options]\n");
  printf("Options:\n");
  printf("  --help                          Display this information and exit\n");
  printf("  --version                       Display the version number and exit\n");
  printf("  geno_file=<file>                VCF or TPED file\n");
  printf("                                  No default value, must be specified\n");
  printf("  geno_gz=<0 or 1>                Set to 1 if geno_file is compressed with gzip.\n");
  printf("                                  The default is determined by the file extension.\n");
  printf("  --notGroupedByChr               If geno_file is a TPED file and it is not grouped by chromosome\n");
  printf("  sortedByLoc=<0 or 1>            If geno_file is not sorted by location within each chr, set to 0.\n");
  printf("  allele_miss=<character>         Missing value for alleles in a TPED file\n");
  printf("                                  The default is 0\n");
  printf("  geno_sep=<character>            File delimiter for the genotype file. Use t or 't' for tab-delimited files.\n");
  printf("                                  The default is a space (' ') for TPED files and a tab for VCF files.\n");
  printf("  max_maf=<value>                 Maximum MAF to use for which SNPs to include in the analysis\n");
  printf("                                  Any SNP with MAF > max_maf will be ignored\n");
  printf("                                  The default value is 0.5\n");
  printf("  min_maf=<value>                 Minimum MAF to use for which SNPs to include in the analysis\n");
  printf("                                  Any SNP with MAF < min_maf will be ignored\n");
  printf("                                  The default value is 0\n");
  printf("  max_miss_rate=<value>           Maximum missing rate for any SNP to be included in the analysis\n");
  printf("                                  The default value is 0.8\n");
  printf("  include_chr=<list>              List of chromsomes to include in the analysis.\n");
  printf("                                  Examples: 1,4,6  1-11,22,  1,3,5-10,X\n");
  printf("                                  The default value is 1-26\n");
  printf("  exclude_chr=<list>              List of chromsomes to exclude in the analysis.\n");
  printf("                                  Examples: 1,4,6  1-11,22,  1,3,5-10,X\n");
  printf("  include_snp=<file>              File containing 1 column of the SNPs to include in the analysis.\n");
  printf("  exclude_snp=<file>              File containing 1 column of the SNPs to exclude in the analysis.\n");
  printf("  include_gene=<file>             File containing 1 column of the genes to include in the analysis.\n");
  printf("  exclude_gene=<file>             File containing 1 column of the genes to exclude in the analysis.\n");
  printf("  chr_alias_file=<file>           Tab delimited text file with a chromosome number (integer 1-26) in column 1,\n");
  printf("                                  and chromosome string in column 2.\n");



  printf("  pheno_file=<file>               Uncompressed file containing the response and covariates\n");
  printf("                                  No default value, must be specified\n");
  printf("  pheno_miss=<character>          Missing value for the pheno_file\n");
  printf("                                  The default value is '.'\n");
  printf("  covars=<List of variables>      Comma separated list of variable names or column numbers in pheno_file to use as covariates in the analysis.\n");
  printf("                                  Example: covars=4,7,9 will use columns 4,7 and 9 in pheno_file as covariates.\n");
  printf("                                  The default value is to not use covariates.\n");
  printf("  pheno_sep=<character>           One character value for the phenotype file delimiter\n");
  printf("                                  The default value is a tab ('t')\n");
  printf("  id=<variable>                   Variable name or column number for the subject ids\n");
  printf("                                  The default value is 1\n");
  printf("  response=<variable>             Variable name or column number for the response\n");
  printf("                                  The default value is 2\n");
  printf("  pheno_header=<0 or 1>           Set to 1 if row 1 of pheno_file contains column names.\n");
  printf("                                  The default value is determined from the file.\n");
  printf("  case_value=<integer>            Value representing cases if response_col is for a case-control study.\n");
  printf("                                  The default value is 1.\n");
  printf("  control_value=<integer>         Value representing controls if response_col is for a case-control study.\n");
  printf("                                  The default value is 0.\n");
  printf("  include_sub=<file>              File containing 1 column of the subject ids to include in the analysis.\n");
  printf("  exclude_sub=<file>              File containing 1 column of the subject ids to exclude in the analysis.\n");
  printf("  --useAllSubsForV                Option to use both cases and controls for the computation of each genotype covariance matrix.\n");
  printf("                                  Otherwise only the controls will be used. This option is only valid for case-control studies.\n");

  printf("  out_gene=<file>                 Output file for the genes\n");
  printf("                                  The default value is ./out_gene.txt.\n");
  printf("  out_snp=<file>                  Output file for the SNPs\n");
  printf("                                  The default value is ./out_snp.txt.\n");

  printf("  gene_file=<dir>                 Folder containing the gene information files\n");
  printf("                                  The default value is hg19.\n");
  printf("  reflink.geneName=<string>       Name of the variable in the INFO field of the VCF file that gives the gene name\n");
  printf("                                  The default value is reflink.geneName\n");
  printf("  --no_gene_file                  Do not use the default gene table file for the gene names.\n");
  printf("                                  Instead use the reflink.geneName value in the VCF file.\n");
  printf("                                  This option is only valid if geno_file is a VCF file and if the weight= option is not specified.\n");
  
  printf("  weight_file=<file>              Uncompressed file containing the SNP weights. This file may have 1 or 2 columns only.\n");
  printf("                                  For a 1 column file the column of weights must match the order of the SNPs in geno_file.\n");
  printf("                                  For a 2 column file, the first column is the SNP id and the second column the weight.\n");
  printf("  weight_sep=<character>          One character value for the weight file delimiter.\n");
  printf("  weight_header=<0 or 1>          Set to 1 if row 1 of weight_file contains column names.\n");
  printf("                                  The default value is determined from the file.\n");
  printf("  weight=<c,e,ns>                 Includes only a particular subset of SNPs in the analysis.\n");
  printf("                                  Set to c to only include codon SNPs, e to only include exons, and ns for non-synonomous SNPs.\n");
  printf("                                  The default is to include all SNPs\n");

  printf("  tests=<List of strings>         Comma delimited list of tests to use. The valid tests are:\n");
  printf("                                  ALL, BM, CAL, CMC, DAS, FI, HOI, HOT, MDF, MDP, RO, RV, SKAT, STO, STP, SUM, TH, C1-C5\n");
  printf("                                  The strings are not case-sensitive.\n");
  printf("                                  Example: tests=st,bm,sum will compute the SUM, Stouffer, and BM tests.\n");
  printf("                                  Example: tests=c1,th will compute the user defined test c1 and the threshold test.\n");
  printf("                                  Use tests=all to compute all non user defined tests.\n"); 
  printf("                                  The default is to use the Rover test\n");
  printf("  nperm=<integer>                 Number of permutations in each iteration of the adaptive permutation algorithm. See adapt_maxperm.\n");
  printf("                                  The default value is 500\n");
  printf("  seed=<integer>                  Initial seed for the random permutations\n");
  printf("                                  The default is the returned value from the time() function\n");
  printf("  adapt_maxperm=<integer>         Maximum number of adaptive permutations. Set to a value > 2*nperm for adaptive permutations.\n");
  printf("                                  The stopping criteria is checked after every nperm permutations. See nperm.\n");
  printf("                                  The default value is 1000000\n");
  printf("  adapt_tol=<value>               Stopping tolerance for adaptive permutations\n");
  printf("                                  The default is 0.1\n");

  printf("  das_alpha=<value>               A number between 0 and 1 for the alpha parameter in the Data Adaptive Sum Test\n");
  printf("                                  The default is 0.05\n");
  printf("  threshold_alpha=<value>         A number between 0 and 1 for the alpha parameter in the Theshold Test\n");
  printf("                                  The default is 0.05\n");
  printf("  bm_n=<integer>                  Number of points to consider in the Bounded Minimum Test.\n");
  printf("                                  The default is 10\n");
  
  printf("  rover_t=<value>                 Positive number t to define the weights (w = sqrt(1-exp(-t*delta*delta))) in the Rover Test.\n");
  printf("                                  The default is 0.2\n");
  printf("  vs=<value>                      The value of vs, where vs abbreviates variance stabilization, is the value added to the estimated variance of each SNP.\n");
  printf("                                  The default is 0.005\n");

  printf("  print=<integer>                 The larger the integer, the more information will be printed\n");
  printf("                                  The default is 0 (no printing)\n");

  printf("  update_cor=<value>              Value such that the correlation matrix will be updated if the relative difference in the observed and\n");
  printf("                                  permuted Hotelling test statistics is less than update_cor.\n");
  printf("                                  This option is only valid for case-control data and when the option --useAllSubsForV is not used.\n");
  printf("                                  The default is 0.3\n");
  printf("  addcols=<0,1,2>                 If 1, then the number of permutations for each test will be output in the out_gene file.\n");
  printf("                                  If 2, then the number of permutations and the number of times the permutation test statistic was greater\n");
  printf("                                  than the observed test statistic for each test wil be output in the out_gene file.\n");
  printf("                                  The default is 0.\n");

  printf("  gene_kb=<integer>               Number of kilobases upstream and downstream to expand the gene regions in the gene tables.\n");
  printf("                                  The default is 10.\n");
  printf("  store_perms=<0 or 1>            Flag to store the permutations in a matrix that will be used for every gene.\n");
  printf("                                  Should only be used for small sample sizes and number of permutations. The default is 0.\n");
  

  /*
  printf("  refseq.changesAA=<string>       Name of the variable in the INFO field of the VCF file that gives the value of refseq.changesAA\n");
  printf("                                  The default value is refseq.changesAA\n");
  printf("  refseq.functionalClass=<string> Name of the variable in the INFO field of the VCF file that gives the value of refseq.functionalClass\n");
  printf("                                  The default value is refseq.functionalClass\n");
  printf("  gene_buffer=<integer>           Maximum number of characters for each value the reflink.geneName variable\n");
  printf("                                  The default value is 100\n");
  printf("  chr_buffer=<integer>            Maximum number of characters of any value in the CHR field of the VCF file\n");
  printf("                                  The default value is 20\n");
  printf("  id_buffer=<integer>             Maximum number of characters of any id in the phenotype and genotype files\n");
  printf("                                  The default value is 100\n");
  printf("  pheno_buffer=<integer>          Buffer size for reading each row of the phenotype file\n");
  printf("                                  The default value is the number of characters of row 1 + 1000\n");
  printf("  geno_buffer=<integer>           Buffer size for reading each line of the genotype file\n");
  printf("                                  The default value is the number of characters of the first non-comment row + 500(TPED) or 5000(VCF)\n")
  printf("  --no_snp_in_mult_genes          For not allowing the same SNP to be used in multiple genes\n");
  printf("  maxChrNum=<integer>             Maximum chromsome number in the gene table\n");
  printf("                                  The default is 27\n");
  */

  printf("\n For additional help on a particular option try: crave <option> --help. Example: crave tests= --help\n");



} /* END: print_help */

/* Function to print the info struct */
void print_info(info, which)
INFO *info;
int which;
{
  char *type[3] = {" ", "VCF", "TPED"}, *yesno[2] = {"no", "yes"}, *yy[2] = {"continuous", "binary"};

  switch (which) {
    case 1:
      /* Call after init */
      if (FPT_LOG_FILE != NULL) {
        printf_s("log file:                                   %s\n", info->log_file);
      }
      printf_s("output gene file:                           %s\n", info->out_file);
      printf_s("output snp file:                            %s\n", info->out_snp);
      printf_s("genotype file:                              %s\n", info->geno_file);
      printf_s("genotype file type:                         %s\n", type[info->file_type]);
      printf_s("genotype file compressed:                   %s\n", yesno[(int) info->gfpt.gzFlag]);
      printf_s("genotype file grouped by chromosome?:       %s\n", yesno[(int) info->groupedByChr]);
      printf_i("genotype file buffer size:                  %d\n", (int) info->geno_buffer);
      if (info->file_type == TPED) {
        printf_c("missing allele character:                   %c\n", info->allele_miss);
      }
      if (info->inc_snp_flag) {
        printf_s("included SNP file:                          %s\n", info->snp_file);
      }
      if (info->exc_snp_flag) {
        printf_s("excluded SNP file:                          %s\n", info->snp_file);
      }

      printf_s("phenotype file:                             %s\n", info->pheno_file); 
      printf_i("phenotype file buffer size:                 %d\n", (int) info->pheno_buffer);
      printf_i("phenotype file number of rows:              %d\n", info->pheno_nrow);
      printf_i("phenotype file number of columns:           %d\n", info->pheno_ncol);
      printf_c("phenotype file missing value character:     %c\n", info->pheno_miss);
      printf_s("phenotype file header of variable names:    %s\n", yesno[info->pheno_header]);
      printf_i("phenotype file response column:             %d\n", info->response_col);
      if (info->response_var != NULL) {
        printf_s("response variable:                          %s\n", info->response_var);
      }
      if (info->id_col > 0) printf_i("phenotype file id column:                   %d\n", info->id_col);
      if (info->id_var != NULL) {
        printf_s("id variable:                                %s\n", info->id_var);
      }
      if (info->inc_sub_flag) {
        printf_s("included subject file:                      %s\n", info->sub_file);
      }
      if (info->exc_sub_flag) {
        printf_s("excluded subject file:                      %s\n", info->sub_file);
      }


      if (info->useGeneTable) {
        printf_s("gene table file:                            %s\n", info->gene_file);
      }
      if (info->inc_gene_flag) {
        printf_s("included gene file:                         %s\n", info->genelist_file);
      }
      if (info->exc_gene_flag) {
        printf_s("excluded gene file:                         %s\n", info->genelist_file);
      }

      if (info->weight_file) {
        printf_s("SNP weight file:                            %s\n", info->weight_file); 
      }

      printf_i("number of tests:                            %d\n", info->n_test);
      printf_i("number of permutations:                     %d\n", info->nperm);
      printf_s("adaptive permutations:                      %s\n", yesno[info->adapt_flag]);
      if (info->adapt_flag) {
        printf_i("number of adaptive iterations:              %d\n", info->adapt_maxiter);
        printf_d("adaptive tolerance:                         %g\n", info->adapt_tol);
      }

      printf_d("minimum MAF:                                %g\n", info->min_maf);
      printf_d("maximum MAF:                                %g\n", info->max_maf);
      printf_d("maximum missing rate:                       %g\n", info->max_missRate);
      if (info->chrStr == NULL) {
        printf_s("chromosomes to include:                     %s\n", "1-26");
      } else if (info->inc_chr_flag) {
        printf_s("chromosomes to include:                     %s\n", info->chrStr);
      } else if (info->exc_chr_flag) {
        printf_s("chromosomes to exclude:                     %s\n", info->chrStr);
      } 

      printf_i("maximum length of id strings:               %d\n", info->id_len);
      printf_i("maximum length of chromosome strings:       %d\n", info->chr_size);
      printf_i("maximum length of gene strings:             %d\n", info->gene_size);
      if (!info->useGeneTable) {
        printf_s("reflink.geneName:                           %s\n", info->gene_var);
        printf_s("refseq.changesAA:                           %s\n", info->changesAA_var);
        printf_s("refseq.functionalClass:                     %s\n", info->functionalClass_var);
      }
      printf_i("print:                                      %d\n", info->print);
      /*
      printf_i("allow SNPs to be in multiple genes:         %d\n", info->snp_in_mult_genes);
      */
      break;
    case 2:
      /* Call after phenotype file is loaded */
      printf_i("number of valid subjects in phenotype file: %d\n", info->nsubs);
      printf_s("response type:                              %s\n", yy[info->CC_flag]);
      printf_i("number of covariates in analysis:           %d\n", info->n_use_covars);
      if (info->CC_flag) {
        printf_i("number of cases:                            %d\n", info->n_case);
        printf_i("number of controls:                         %d\n", info->n_control);
        printf_i("value for cases:                            %d\n", info->case_value);
        printf_i("value for controls:                         %d\n", info->control_value);
      }
      break;
    case 3:
      /* Call after genotype columns are checked */
      printf_i("genotype file number of columns:            %d\n", info->geno_ncols);
      printf_i("genotype file number of subjects:           %d\n", info->n_genotypes);
      printf_i("genotype file starting row number:          %d\n", info->geno_line_num+1);
      break;
    case 4:
      /* Call after update info */
      printf_i("number of subjects used in analysis:        %d\n", info->nsubs);
      if (info->CC_flag) {
        printf_i("number of cases used in analysis:           %d\n", info->n_case);
        printf_i("number of controls used in analysis:        %d\n", info->n_control);
      }
      break;
    default:
      break;
  } 

} /* END: print_info */

/* Function to print a snp link */
void print_snpLink(str, ngeno, snpNum)
SNP_STR *str;
int ngeno, snpNum;
{

  /*
  nmiss = genoCounts(str->G, ngeno, counts);
  printf("SNP %d: Weight=%g, SNP_0=%d, SNP_1=%d, SNP_2=%d, SNP_MISS=%d\n", snpNum, str->weight, counts[0], counts[1], counts[2], nmiss);
  if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "SNP %d: Weight=%g, SNP_0=%d, SNP_1=%d, SNP_2=%d, SNP_MISS=%d\n", snpNum, str->weight, counts[0], counts[1], counts[2], nmiss);
  */
  
  printf("SNP %d: Weight=%g, SNP=%p, SNP->next=%p\n", snpNum, str->weight, str, str->next);
  if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "SNP %d: Weight=%g, SNP=%p, SNP->next=%p\n", snpNum, str->weight, str, str->next);
  

  /*print_gvec(str->G, ngeno, "genotypes");*/

} /* END: print_snpLink */

void print_geneLink(str, ngeno, geneNum)
GENE_STR *str;
int ngeno, geneNum;
{
  int i;
  SNP_STR *snp;

  printf("GENE %d Gene name = %s, Number of SNPs = %d\n", geneNum, str->name, str->n_snps);
  if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "GENE %d Gene name = %s, Number of SNPs = %d\n", geneNum, str->name, str->n_snps);

  snp = str->first;
  for (i=0; i<str->n_snps; i++) {
    print_snpLink(snp, ngeno, i+1);
    snp = snp->next;
  }

} /* END: print_geneLink */

void print_chrStr(info)
INFO *info;
{
  int i;
  CHR_STR *chr=info->chr[info->chr_str_index];
  GENE_STR *gene;

  if (info->useGeneTable) {
    printf_i("### Begin: Chromosome %d ###\n", info->last_chr_num);
  } else {
    printf_s("### Begin: Chromosome %s ###\n", info->last_chr);
  }
  printf_i("Number of genes = %d\n", chr->n_genes);
  gene = chr->first;
  if (!gene) return;
  for (i=0; i<chr->n_genes; i++) {
    print_geneLink(gene, info->nsubs, i+1);
    gene = gene->next;
    printf_s("\n", NULL);
  }
  if (info->useGeneTable) {
    printf_i("### End: Chromosome %d ###\n", info->last_chr_num);
  } else {
    printf_s("### End: Chromosome %s ###\n", info->last_chr);
  }

} /* END: print_chrStr */

static void print_R(str0, str, pdbl, pint, pchar)
char *str0, *str;
double *pdbl;
int *pint;
char *pchar;
{
   #if RFLAG
   print_r_main(str0, str, pdbl, pint, pchar);
   #endif

} /* END: print_R */

void printf_s(str, arg)
char *str;
char *arg;
{
 
  if (arg == NULL) {
    printf(str);
    if (FPT_LOG_FILE) {
      fprintf(FPT_LOG_FILE, str);
      fflush(FPT_LOG_FILE);
    }
  } else {
    printf(str, arg);
    if (FPT_LOG_FILE) {
      fprintf(FPT_LOG_FILE, str, arg);
      fflush(FPT_LOG_FILE);
    }
  } 
  if (RFLAG) print_R(str, arg, NULL, NULL, NULL);


} /* END: printf_s */

void printf_i(str, arg)
char *str;
int arg;
{
  printf(str, arg);
  if (FPT_LOG_FILE) {
    fprintf(FPT_LOG_FILE, str, arg);
    fflush(FPT_LOG_FILE);
  } 
  if (RFLAG) print_R(str, NULL, NULL, &arg, NULL);

} /* END: printf_i */

void printf_d(str, arg)
char *str;
double arg;
{
  printf(str, arg);
  if (FPT_LOG_FILE) {
    fprintf(FPT_LOG_FILE, str, arg);
    fflush(FPT_LOG_FILE);
  } 
  if (RFLAG) print_R(str, NULL, &arg, NULL, NULL);

} /* END: printf_d */

void printf_c(str, arg)
char *str;
char arg;
{
  printf(str, arg);
  if (FPT_LOG_FILE) {
    fprintf(FPT_LOG_FILE, str, arg);
    fflush(FPT_LOG_FILE);
  }
  if (RFLAG) print_R(str, NULL, NULL, NULL, &arg);
  
} /* END: printf_c */



