/***************************************************************
                                                             
                    CRAVE                  
                                                              
 This file is distributed under the GNU General Public        
 License, Version 2. See the file COPYING.txt for more    
 details                                                      
                                                              
****************************************************************/

#include "./crave.h" 

/**************************************************************************************
************************* Normal ******************************************************
***************************************************************************************/ 

/* Normal pdf */
double normal_pdf(x, mu, sigma2)
double x, mu, sigma2;
{
  double ret, tmp1, tmp2;

  tmp1 = 2*sigma2;
  tmp2 = x - mu;
  ret = exp(-tmp2*tmp2/tmp1)/sqrt(MYPI*tmp1);
  return(ret);

} /* END: normal_pdf */

/* Normal N(0, 1) pdf */
double normal01_pdf(x)
double x;
{
  double ret;

  ret = exp(-0.5*x*x)*ONE_OVER_SQRT_2PI;
  return(ret);

} /* END: normal01_pdf */

/* Normal cdf for x > 0 */
double normal_cdf(x, mu, sigma2)
double x, mu, sigma2;
{
  double b0=0.2316419, b1=0.319381530, b2=-0.356563782, b3=1.781477937, b4=-1.821255978, b5=1.330274429;
  double z, sum, t, ret, t2, t3, t4, t5, pdf;

  z = fabs(x-mu)/sqrt(sigma2);

  pdf = normal01_pdf(z);
  t = 1.0/(1.0 + b0*z);
  sum = b1*t;
  t2 = t*t;
  sum += b2*t2;
  t3 = t2*t;
  sum += b3*t3;
  t4 = t3*t;
  sum += b4*t4;
  t5 = t4*t;
  sum += b5*t5; 

  ret = 1.0 - pdf*sum;

  return(ret);

} /* END: normal_pdf */

/* Normal(0,1) cdf for x > 0 */
static double normal01_cdf(z)
double z;
{
  double b0=0.2316419, b1=0.319381530, b2=-0.356563782, b3=1.781477937, b4=-1.821255978, b5=1.330274429;
  double sum, t, ret, t2, t3, t4, t5, pdf;

  z = fabs(z);
  pdf = normal01_pdf(z);
  t = 1.0/(1.0 + b0*z);
  sum = b1*t;
  t2 = t*t;
  sum += b2*t2;
  t3 = t2*t;
  sum += b3*t3;
  t4 = t3*t;
  sum += b4*t4;
  t5 = t4*t;
  sum += b5*t5; 

  ret = 1.0 - pdf*sum;

  return(ret);

} /* END: normal01_pdf */

/* Function to compute the inverse error function */
double erf_inv(x, error)
double x;
int *error;
{
  double ret, temp, ltemp, arg1;

  *error = 0;
  temp   = 1.0 - x*x;
  ltemp  = log(temp);
  arg1   = ERF_2_PIA + ltemp/2.0;
  ret    = sqrt(sqrt(arg1*arg1 - ltemp/ERF_A) - arg1);
  if (x < 0) ret = -ret;

  return(ret);

} /* END: erf_inv */

/* Function to compute the inverse normal cdf */
double normal01_cdfinv(x)
double x;
{
  double ret;
  int error;

  /* First check for a standard value of x */
  if (fabs(x - 0.05) < ALMOST_ZERO) {
    return(N01_CDFINV_05);
  } else if (fabs(x - 0.95) < ALMOST_ZERO) {
    return(N01_CDFINV_95);
  } else if (fabs(x - 0.01) < ALMOST_ZERO) {
    return(N01_CDFINV_01);
  } else if (fabs(x - 0.99) < ALMOST_ZERO) {
    return(N01_CDFINV_99);
  } else if (fabs(x - 0.10) < ALMOST_ZERO) {
    return(N01_CDFINV_10);
  } else if (fabs(x - 0.90) < ALMOST_ZERO) {
    return(N01_CDFINV_90);
  } else {
    ret = erf_inv(2.0*x-1.0, &error);
    ret = SQRT2*ret;
    return(ret);
  }

} /* normal01_cdfinv */

/* Function to compute a normal 2-sided p-value */
double pvalue_normal(x, mu, sigma2)
double x, mu, sigma2;
{
  double pval;

  pval = 2.0*(1.0 - normal_cdf(x, mu, sigma2));
  pval = MIN(pval, 1.0);

  return(pval);

} /* END: pvalue_normal */

/* Function to compute a normal(0,1) 2-sided p-value */
double pvalue_normal01(x)
double x;
{
  double pval;

  pval = 2.0*(1.0 - normal01_cdf(x));
  pval = MIN(pval, 1.0);

  return(pval);

} /* END: pvalue_normal */

/* Function to compute Y - meanY */
void YminusYbar(Y, n, meanY, ret)
yType *Y;
int n;
double meanY, *ret;
{
  int i;
  yType *pb;
  double *pd;

  for (i=0, pd=ret, pb=Y; i<n; i++, pd++, pb++) *pd = *pb - meanY;
  
} /* YminusYbar */

/* Function to compute the mean and variance of a double vector */
double mean_var_dvec(vec, n, variance)
double *vec;
int n;
double *variance;
{
  double sum=0.0, sum2=0.0, ret, *p, temp;
  int i;

  for (i=0, p=vec; i<n; i++, p++) {
    temp = *p;
    sum += temp;
    sum2 += temp*temp;
  }
  ret = sum/n;
  *variance = (sum2 - n*ret*ret)/(n - 1.0);

  return(ret);

} /* END: mean_var_dvec */

/* Function to compute the mean and variance of each column of a matrix */
void mean_var_dmat(mat, nr, nc, mean, variance)
double **mat;
int nr, nc;
double *mean, *variance;
{
  double sum, sum2, temp;
  int i, j;

  for (j=0; j<nc; j++) {
    sum  = 0.0;
    sum2 = 0.0;
    for (i=0; i<nr; i++) {
      temp = mat[i][j];
      sum += temp;
      sum2 += temp*temp;
    }
    temp = sum/nr;
    mean[j] = temp;
    variance[j] = (sum2 - nr*temp*temp)/(nr - 1.0);
  }

} /* END: mean_var_dmat */


/* Function to compute the genotype frequency counts for a SNP.
    Missing count is returned */
int genoCounts(SNP, n, ret)
genoType *SNP;
int n, *ret; /* ret is vector of length 3 */
{
  int i, nmiss=0, geno;
  genoType *pg;

  ret[0] = ret[1] = ret[2] = 0;
  for (i=0, pg=SNP; i<n; i++, pg++) {
    geno = (int) *pg;
    if (geno != MISS_GENO) {
      ret[geno]++;
    } else {
      nmiss++;
    }
  }

  return(nmiss);

} /* END: genoCounts */

/* Function to permute a double vector */
void permute_dVec(vec, n, outvec)
double *vec, *outvec;
int n;
{
  int i, rnum;
  double temp, *p1, *p2;

  /* If outvec != NULL, copy vec to outvec first */
  if (outvec == NULL) {
    p1 = vec;
  } else {
    for (i=0, p1=vec, p2=outvec; i<n; i++, p1++, p2++) *p2 = *p1;
    p1 = outvec;
  }

  for (i=0; i<n; i++) {
    rnum = rand() % n;
    temp = p1[i];
    p1[i] = p1[rnum];
    p1[rnum] = temp;
  }

} /* END: permute_dVec */

/* Function to get a random permutation of integers from 0 to n-1 using the rand function */
void get1Permutation_c(n, outvec)
int n, *outvec;
{
  int i, *p, rnum, temp;
  
  for (i=0, p=outvec; i<n; i++, p++) *p = i;

  for (i=0; i<n; i++) {
    rnum = rand() % n;
    temp = outvec[i];
    outvec[i] = outvec[rnum];
    outvec[rnum] = temp;
  }

} /* END: get1Permutation_c */

/* Function to get a matrix of random permutations. Each row is a permutation */
void getNPermutation_c(mat, nperm, ncol)
int **mat, nperm, ncol;
{
  int i;
  
  for (i=0; i<nperm; i++) get1Permutation_c(ncol, mat[i]);

} /* END: getNPermutation_c */

/* Function to generate a matrix of permutations */
void getPermMat(info)
INFO *info;
{
  int i, j, nsub=info->nsubs, nperm=info->nperm, rnum;
  double **mat, **prow, *pvec, b1, b2, *vec0, *pvec0, *Ynorm=info->Y_norm, *py;

  if (!info->storePermFlag) return;
  mat  = dblMatrix_alloc(nperm, nsub);
  vec0 = (double *) malloc(nsub*sizeof(double));
  CHECK_MEM(vec0);   
  for (i=0, pvec=vec0, py=Ynorm; i<nsub; i++, pvec++, py++) *pvec = *py; 

  for (j=0, prow=mat; j<nperm; j++, prow++) {
    for (i=0, pvec=vec0; i<nsub; i++, pvec++) {
      rnum       = rand() % nsub;
      b1         = *pvec;
      b2         = vec0[rnum];
      *pvec      = b2;
      vec0[rnum] = b1;
    }
    /* Copy vec0 to a row */
    for (i=0, pvec0=vec0, pvec=*prow; i<nsub; i++, pvec0++, pvec++) *pvec = *pvec0;
  }

  free(vec0);
  info->storePermMat = mat;

} /* END: getPermMat */

/* Function to compute the mean of a genotype vector */
static double snp_mean(snp, n, ccFlag, controls)
genoType *snp;
int n;
binary ccFlag, *controls;
{
  int i, temp, count=0;
  double sum=0.0;
  genoType *p;
  binary *pb;

  if (ccFlag) {
    for (i=0, p=snp, pb=controls; i<n; i++, p++, pb++) {
      if (*pb) {
        temp = (int) *p;
        if (temp != MISS_GENO) {
          sum += temp;
          count++;
        }
      } /* END: if (*py < 0.5) */
    }
    if (!count) printf_s("\n***ERROR***: no controls with non-missing genotypes to compute the SNP mean\n", NULL);
  } else {
    for (i=0, p=snp; i<n; i++, p++) {
      temp = (int) *p;
      if (temp != MISS_GENO) {
        sum += temp;
        count++;
      }
    }
    if (!count) printf_s("\n***ERROR***: no subjects with non-missing genotypes to compute the SNP mean\n", NULL);
  }
  if (count) {
    sum = sum/count;
  } else {
    sum = DOUBLE_MISS;
  }

  return(sum);

} /* END: snp_mean */

/* Function to compute the mean of the snps in a gene */
int gene_mean(GENE, nsub, mean, ccFlag, controls)
GENE_STR *GENE;
int nsub;
double *mean; /* Return vector of length GENE->n_snps */
binary ccFlag, *controls;
{
  SNP_STR *SNP;
  int i, ret=0;
  double *pd, errorVal=DOUBLE_MISS+0.5, temp;

  SNP = GENE->first;
  for (i=0, pd=mean; i<GENE->n_snps; i++, pd++) {
    temp = snp_mean(SNP->G, nsub, ccFlag, controls);
    *pd = temp;
    if (temp < errorVal) ret = ERROR_NO_NONMISS_GENO;
    SNP = SNP->next; 
  }

  return(ret);

} /* END: gene_mean */

/* Function to determine if any snps in a gene are monomorphic */
static int getMonoSNPs(GENE, nsub, controls, ret, flag)
GENE_STR *GENE;
int nsub;
binary *controls, *ret;
int flag; /* Set to 0 if monomorphic snps are known as in computing invcor */
{
  int i, j, nmono, nsnps=GENE->n_snps;
  binary *pb, *pret;
  SNP_STR *SNP;
  genoType *psnp;

  if (flag) {
    nmono = nsnps;
  } else {
    nmono = 0;
  }
  SNP = GENE->first;

  for (i=0, pret=ret; i<nsnps; i++, pret++) {
    if (flag) {

      /* Assume snp is monomorphic */
      *pret = 1;
      SNP->mono_control = 1;
      for (j=0, psnp=SNP->G, pb=controls; j<nsub; j++, psnp++, pb++) {
        /* Only use the controls */
        if ((*pb) && (*psnp > 0)) {
          *pret = 0;
          SNP->mono_control = 0;
          nmono--;
          break;
        }
      }
    } else {
      j = SNP->mono_control;
      *pret = j;
      if (j) nmono++;
    }
    SNP = SNP->next;
  }

  return(nmono);

} /* getMonoSNPs */

/* Function to compute the covariance matrix V of snps in a gene */
static int gene_cov_diag(GENE, nsub, cov, ccFlag, controls, ncontrol)
GENE_STR *GENE;
int nsub;
double **cov; /* Return matrix of length GENE->n_snps */
binary ccFlag, *controls;
int ncontrol;
{
  SNP_STR *SNP;
  int i, j, nsnp=GENE->n_snps, nuse, n;
  double scale, sum, sum2, mean;
  genoType geno, *psnp;
  binary *pb;

  if (ccFlag) {
    nuse = ncontrol;
  } else {
    nuse = nsub;
  }
  scale = 1.0/(nuse - 1.0);

  SNP = GENE->first;
  for (j=0; j<nsnp; j++) {
    sum  = 0.0;
    sum2 = 0.0;
    n    = 0;
    if (ccFlag) {
      for (i=0, psnp=SNP->G, pb=controls; i<nsub; i++, psnp++, pb++) {
        /* Only use the controls */
        if (*pb) {
          geno = *psnp;
          if (geno != MISS_GENO) {
            sum += geno;
            sum2 += geno*geno;
            n++;
          } 
        }
      }
    } else {
      for (i=0, psnp=SNP->G; i<nsub; i++, psnp++) {
        geno = *psnp;
        if (geno != MISS_GENO) {
          sum += geno;
          sum2 += geno*geno;
          n++;
        } 
      }
    }
    if (!n) return(ERROR_VARIANCE);
    mean = sum/n;
    cov[j][j] = (sum2 - mean*mean*n)*scale;

    SNP = SNP->next;
  }
 
  return(0);
  
} /* END: gene_cov_diag */

/* Function to compute the covariance matrix V of snps in a gene */
static int gene_cov0(GENE, nsub, cov, ccFlag, controls, ncontrol, diagFlag)
GENE_STR *GENE;
int nsub;
double **cov; /* Return matrix of length GENE->n_snps */
binary ccFlag, *controls;
int ncontrol;
binary diagFlag;
{
  SNP_STR *SNP;
  int i, j, nsnp=GENE->n_snps, nuse;
  double **temp, mj, *p, scale, *mean;
  genoType geno, *psnp;
  binary *pb;

  if (diagFlag) return(gene_cov_diag(GENE, nsub, cov, ccFlag, controls, ncontrol));

  if (ccFlag) {
    nuse = ncontrol;
  } else {
    nuse = nsub;
  }

  /* Allocate memory */
  mean = (double *) malloc(nsnp*sizeof(double));
  CHECK_MEM(mean);
  temp = dblMatrix_alloc(nsnp, nuse);

  /* Get the means */
  i = gene_mean(GENE, nsub, mean, ccFlag, controls);
  if (i) {
    /* Problem with missing genotypes */
    matrix_free((void **) temp, nsnp);
    free(mean);
    return(i);
  }

  /* Store G - mu (K x nr) in temp */
  SNP = GENE->first;
  for (j=0; j<nsnp; j++) {
    mj = -mean[j];
    if (ccFlag) {
      for (i=0, psnp=SNP->G, p=temp[j], pb=controls; i<nsub; i++, psnp++, pb++) {
        /* Only use the controls */
        if (*pb) {
          geno = *psnp;
          if (geno != MISS_GENO) {
            *p++ = mj + geno;
          } else {
            *p++ = DOUBLE_MISS;
          }
        }
      }
    } else {
      for (i=0, psnp=SNP->G, p=temp[j]; i<nsub; i++, psnp++, p++) {
        geno = *psnp;
        if (geno != MISS_GENO) {
          *p = mj + geno;
        } else {
          *p = DOUBLE_MISS;
        }
      }
    }
    SNP = SNP->next;
  }
  free(mean);

  /* Compute (G-mu)(G-mu)` */
  X_Xp_miss(temp, nsnp, nuse, DOUBLE_MISS, cov);

  matrix_free((void **) temp, nsnp);

  /* Check the variances */
  for (i=0; i<nsnp; i++) {
    if (cov[i][i] < NUM_ZERO) return(ERROR_VARIANCE);
  }

  /* Divide by the number of subs */
  scale = 1.0/(nuse - 1.0);
  for (i=0; i<nsnp; i++) {
    for (j=0, p=cov[i]; j<nsnp; j++, p++) *p = *p * scale;
  }

  return(0);
  
} /* END: gene_cov0 */

/* Function to compute an inverse correlation matrix from a covariance matrix */
static int getInvCor(cov, n, invcor, cor, info)
double **cov;
int n;
double **invcor; /* Return matrix */
double **cor; /* NULL or return correlation matrix */
INFO *info;
{
  double **cor2, **pcor, add, mul, temp;  
  int i, flag, j, k;

  if (cor == NULL) {
    cor2 = dblMatrix_alloc(n, n);
    pcor = cor2;
    flag = 0;
  } else {
    flag = 1;
    pcor = cor;
  }

  /* Compute the covariance matrix */
  DV_V_DV(cov, n, pcor);

  /* Compute the inverse cor matrix */
  i = cov_inv(pcor, n, invcor);

  /* Alter the correlation matrix if singular */
  if (i) {
    if (info->print > PRINT_INVCORR) printf_s("***WARNING*** Correlation matrix is singular\n", NULL);
    add = info->INV_ADD;
    mul = 1.0/info->INV_DIV;
    for (j=0; j<n; j++) {
      pcor[j][j] += add;  
      for (k=j; k<n; k++) {
        temp = pcor[j][k]*mul;
        pcor[j][k] = temp;
        pcor[k][j] = temp;
      }
    } 
    i = cov_inv(pcor, n, invcor);
    if ((!i) &&  (info->print > PRINT_INVCORR)) printf_s("***NOTE*** Adjusted correlation matrix is non-singular\n", NULL);

  } /* END: if (i) */

  if (!flag) matrix_free((void **) cor2, n);
  if (i) return(ERROR_SINGULAR_MATRIX);

  return(0);

} /* END: getInvCor */

/* Function to compute the covariance matrix V and inverse correlation matrix of snps in a gene */
static int gene_cov_invcor(GENE, nsub, cov, ccFlag, controls, ncontrol, invFlag, invcor, retInvFlag, cor, info, snpFreqFlag)
GENE_STR *GENE;
int nsub;
double **cov; /* Return covariance matrix of length GENE->n_snps */
binary ccFlag, *controls;
int ncontrol;
binary invFlag; /* Flag to compute the inverse correlation matrix */
double **invcor;
binary *retInvFlag; /* return flag for the inverse corr matrix */
double **cor; /* Use NULL for not retuning the correlation matrix */
INFO *info;
int snpFreqFlag;
{
  int i, nsnp=GENE->n_snps, nmono=0, nsnp1;
  binary *monoSNP=NULL, corFlag, diagFlag;
  GENE_STR *GENE1=NULL;
  double **cov1=NULL, **cor1=NULL, **cor2=NULL;

  *retInvFlag = 1;
  if (cor == NULL) {
    corFlag = 0;
  } else {
    corFlag = 1;
  }

  diagFlag = (info->geno_cov_flag == 0) ? 1 : 0;

  /* First, determine if there are any monomorphic snps among the controls.
     If there are, keep track of these snps */
  if (ccFlag) {
    monoSNP = (binary *) malloc(nsnp*sizeof(binary));
    CHECK_MEM(monoSNP);
    nmono = getMonoSNPs(GENE, nsub, controls, monoSNP, 1);
    nsnp1 = nsnp - nmono;
    GENE->n_mono_control = nmono;
    if (nmono) {
      /* Get the covariance matrix using all subs */
      ccFlag = 0;
      i = gene_cov0(GENE, nsub, cov, ccFlag, controls, ncontrol, diagFlag);
      if ((i) || (!nsnp1)) {
        /* All subs failed or all snps are monomorphic, so return */
        free(monoSNP);
     
        /* Compute inverse correlation matrix */
        if ((!i) && (invFlag)) *retInvFlag = getInvCor(cov, nsnp, invcor, cor, info); 
        return(i);
      }

      /* Try computing the inverse correlation matrix using all subs */
      if (invFlag) {
        *retInvFlag = getInvCor(cov, nsnp, invcor, cor, info);     
        if (*retInvFlag) invFlag = 0;
      } 

      /* Create a new gene structure excluding the monomorphic snps */
      GENE1 = (GENE_STR *) malloc(sizeof(GENE_STR));
      CHECK_MEM(GENE1);

      copyGeneStr(GENE, monoSNP, nsub, GENE1, snpFreqFlag);

      nsnp1 = GENE1->n_snps;

      /* Compute the cov matrix using only the controls */
      cov1 = dblMatrix_alloc(nsnp1, nsnp1);
      ccFlag = 1;
      i = gene_cov0(GENE1, nsub, cov1, ccFlag, controls, ncontrol, diagFlag);

      if (i) {
        matrix_free((void **) cov1, nsnp1);
        free(monoSNP);
        free_gene(GENE1, 1);
        return(0);
      } else {
        /* Set rows/cols in cov to cov1 */
        copySqMat_exclude(cov, monoSNP, nsnp, cov1, nsnp1);

        /* Compute inverse correlation matrix from controls */
        if (invFlag) {
          cor1 = dblMatrix_alloc(nsnp1, nsnp1);
          if (corFlag) cor2 = dblMatrix_alloc(nsnp1, nsnp1);
          *retInvFlag = getInvCor(cov1, nsnp1, cor1, cor2, info);
          if (!*retInvFlag) copySqMat_exclude(invcor, monoSNP, nsnp, cor1, nsnp1);
          if (corFlag) copySqMat_exclude(cor, monoSNP, nsnp, cor2, nsnp1);
        } 
      }
    } else {
      /* No monomorphic snps, first try using the controls */
      ccFlag = 1;
      i = gene_cov0(GENE, nsub, cov, ccFlag, controls, ncontrol, diagFlag);
      if (i) {
        ccFlag = 0;
        i = gene_cov0(GENE, nsub, cov, ccFlag, controls, ncontrol, diagFlag);
      }
      if ((!i) && (invFlag)) *retInvFlag = getInvCor(cov, nsnp, invcor, cor, info); 
    }
  } else {
    i = gene_cov0(GENE, nsub, cov, ccFlag, controls, ncontrol, diagFlag);
    if ((!i) && (invFlag)) *retInvFlag = getInvCor(cov, nsnp, invcor, cor, info); 
  }

  if (monoSNP != NULL) free(monoSNP);
  if (cov1 != NULL) matrix_free((void **) cov1, GENE1->n_snps);
  if (cor1 != NULL) matrix_free((void **) cor1, GENE1->n_snps);
  if (cor2 != NULL) matrix_free((void **) cor2, GENE1->n_snps);
  if (GENE1 != NULL) free_gene(GENE1, 1);

  return(i);
  
} /* END: gene_cov_invcor */

/* Function to calculate Y_norm */
void getY_norm(info)
INFO *info;
{
  binary ccFlag=info->CC_flag;
  int nsub=info->nsubs, nc, i;
  double *temp, *pd;
  yType *Y=info->CC, *py;
  
  /* Allocate memory for the normalized Y */
  info->Y_norm = (double *) malloc(nsub*sizeof(double));
  CHECK_MEM(info->Y_norm);
  temp = (double *) malloc(nsub*sizeof(double));
  CHECK_MEM(temp);

  nc = info->n_use_covars + 1;

  /* Calculate mu from linear or logistic regression */
  i = EY_X(info->covmat, nsub, nc, Y, ccFlag, temp);
  if (i) {
    printf_s("\nERROR: in computing E[Y|X]. Is the design matrix of full rank?\n", NULL);
    exit(1);
  }

  /* Compute Y1 = Y - mu, and summary stats */
  for (i=0, py=Y, pd=temp; i<nsub; i++, py++, pd++) *pd = *py - *pd;

  /* Normalize */
  norm_dvec(temp, nsub, nsub-1, info->Y_norm); 

  free(temp);

} /* END: getY_norm */

/* Function to compute the deltas given Y */
static void compute_delta(Y, GENE, delta, sdG, wgtFlag)
yType *Y;
GENE_STR *GENE;
double *delta;
double *sdG;  /* sqrt(diag(V)) */
int wgtFlag;
{
  
  SNP_STR *SNP;
  int i, j, n, *pi1;
  double sum, *pd1, sum2, *pd2;

  SNP = GENE->first;
  for (i=0, pd1=delta, pd2=sdG; i<GENE->n_snps; i++, pd1++, pd2++) {
    sum = 0.0;
    n = SNP->n1;
    if (n) {

      for (j=0, pi1=SNP->id1; j<n; j++, pi1++) sum += Y[*pi1];
    } 
    n = SNP->n2;
    if (n) {
      sum2 = 0.0;
      for (j=0, pi1=SNP->id2; j<n; j++, pi1++) sum2 += Y[*pi1];
      sum += sum2 + sum2;
    }

    *pd1 = sum/(*pd2 * SNP->sqrtN);
    if (wgtFlag) *pd1 = *pd1 * SNP->weight;

    SNP = SNP->next; 

  }

} /* END: compute_delta */

/* Function to determine which SNPs have Gik = 2 */
/*
static void find_G_equal2(GENE, nsub)
GENE_STR *GENE;
int nsub;
{
  int i;
  SNP_STR *SNP;
  binary flag=0;

  SNP = GENE->first;
  for (i=0; i<GENE->n_snps; i++) {
    if (SNP->n2) {
      SNP->G_equal_2 = 1;
      flag = 1;
    } else {
      SNP->G_equal_2 = 0;
    }
    SNP = SNP->next;
  }
  GENE->G_equal_2 = flag;

} 
*/

/* Function to compute the sum for Gik=1 in SRT test */
/*
static double srt_sum2(info, GENE, Y)
INFO *info;
GENE_STR *GENE;
double *Y;
{
  int i, j, nsub, *pi1;
  double sum, sum2, one_over_n, temp, *sdG, sdi;
  SNP_STR *SNP;
  genoType *pg;
  
  if (GENE->G_equal_2 == 0) return(0.0);
  
  nsub = info->nsubs;
  sum = 0.0;
  SNP = GENE->first;
  sdG = info->geno_sdG;
  one_over_n = 1.0/nsub;
  for (i=0; i<GENE->n_snps; i++) {
    if (SNP->G_equal_2) {
      sum2 = 0.0;
      pg = SNP->G;
      for (j=0, pi1=SNP->id1; j<SNP->n1; j++, pi1++) sum2 += Y[*pi1];

      temp = sum2*one_over_n;
      sdi = sdG[i];
      sum += sdi*sdi*temp*temp;
    }
    SNP = SNP->next;
  }
  return(sum);  

} 
*/

/* Function to compute the inverse correlation matrix for CMC method */
static void gstar_invcor(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  SNP_STR *SNP, *snp;
  int i, j, nsnp=GENE->n_snps, nsub=info->nsubs, ncat, *cat, *pi, n0, n1, n2, k, k0, nuse, *cat0, cati, catj, freq012[3];
  double **temp, maf, **cor, **cov, **invcor, **mat;
  binary useControls=info->useControlsForV, *controls, retInvFlag;
  genoType *pg0, *pg1, *pg2, *snp0, *snp1, *snp2, *pg, *psnp, geno;
  GENE_STR *GENE1, *pGENE=GENE;

  /* See in info->geno_invcor is NULL */
  if (info->geno_invcor == NULL) {
    if (info->geno_invcor2 == NULL) {
      printf_s("***WARNING***: Inverse correlation matrix is NULL. CMC test will not be computed\n", NULL);
      return;
    }
    pGENE = info->GENE2;
    nsnp  = pGENE->n_snps;
    mat   = info->geno_invcor2;
  } else {
    mat   = info->geno_invcor;
  }

  /* If nsnp = 1, no collapsing */
  if (nsnp == 1) {
    info->gstar_invcor = dblMatrix_alloc(1, 1);
    cor = (info->geno_invcor != NULL) ? info->geno_invcor : info->geno_invcor2;
    info->gstar_invcor[0][0] = cor[0][0];
    info->cmc_ncat = 1;
    info->gstar_ncat = 1; 
    return;
  }

  if (useControls) {
    nuse = info->n_control;
    controls = info->controls;
  } else {
    nuse = nsub;
  }

  cat = (int *) malloc(nsnp*sizeof(int));
  CHECK_MEM(cat);
  for (i=0, pi=cat; i<nsnp; i++, pi++) *pi = 0;

  /* Get the number of categories */
  SNP = pGENE->first;
  i = 3;
  n0 = n1 = n2 = 0;
  for (j=0; j<nsnp; j++) {
    maf = SNP->maf;
    if (maf < 0.02) {
      cat[j] = 0;
      n0++;
    } else if (maf < 0.05) {
      cat[j] = 1;
      n1++;
    } else if (maf < 0.1) {
      cat[j] = 2;
      n2++;
    } else {
      cat[j] = i;
      i++;
    }
    SNP = SNP->next;
  }
  ncat = i; 

  if (!n0) ncat--;
  if (!n1) ncat--;
  if (!n2) ncat--;

  info->cmc_ncat = ncat;
  info->gstar_ncat = ncat; 

  /* If there was no collapsing, then copy invcor to gstar and return */
  if ((!n0) && (!n1) && (!n2)) {
    if (ncat != nsnp) {
      printf_s("\n Internal coding error in gstar_invcor\n", NULL);
      exit(1);
    }
    free(cat);
    temp = info->geno_invcor;
    /* Check that this matrix is not NULL */
    if (temp == NULL) {
      info->errorFlag = 9999;
      return;
    }

    /* Allocate matrix for Gstar and inverse correlation matrix */
    info->gstar_invcor = dblMatrix_alloc(ncat, ncat);

    cor  = info->gstar_invcor;
    for (i=0; i<ncat; i++) {
      for (j=0; j<ncat; j++) cor[i][j] = temp[i][j];
    }
    return;
  }

  /* Save the original categories */
  cat0 = (int *) malloc(nsnp*sizeof(int));
  CHECK_MEM(cat0);
  for (i=0; i<nsnp; i++) cat0[i] = cat[i];

  /* Get the correct category for each snp */
  for (j=0; j<nsnp; j++) {
    k = k0 = cat[j];
    if ((k0 > 0) && (!n0)) k--;
    if ((k0 > 1) && (!n1)) k--;
    if ((k0 > 2) && (!n2)) k--;
    cat[j] = k;
  }

  snp0 = (genoType *) malloc(nsub*sizeof(genoType));
  CHECK_MEM(snp0);
  snp1 = (genoType *) malloc(nsub*sizeof(genoType));
  CHECK_MEM(snp1);
  snp2 = (genoType *) malloc(nsub*sizeof(genoType));
  CHECK_MEM(snp2);
  for (i=0, pg0=snp0, pg1=snp1, pg2=snp2; i<nsub; i++, pg0++, pg1++, pg2++) {
    *pg0 = MISS_GENO;
    *pg1 = MISS_GENO;
    *pg2 = MISS_GENO;
  }

  /* Collapse snps in first 3 cats */
  SNP = pGENE->first;
  for (j=0; j<nsnp; j++) {
    k = cat0[j];

    if (k < 3) {
      if (k == 0) {
        pg = snp0;
      } else if (k == 1) {
        pg = snp1;
      } else {
        pg = snp2;
      }

      for (i=0, psnp=SNP->G; i<nsub; i++, psnp++) {
        geno = *psnp;
        if (geno != MISS_GENO) { 
          if (pg[i] == MISS_GENO) {
            pg[i] = geno;
          } else {
            pg[i] += geno;
          }
        } 
      }
    }
    SNP = SNP->next;
  }

  /* Create a new gene structure from the collapse cats and other snps */
  GENE1 = (GENE_STR *) malloc(sizeof(GENE_STR));
  CHECK_MEM(GENE1);
  GENE1->next = NULL;
  GENE1->previous = NULL;
  GENE1->name = NULL;   
  GENE1->first = NULL;
  GENE1->n_snps = 0;
  if (n0) {
    snp = (SNP_STR *) malloc(sizeof(SNP_STR));
    CHECK_MEM(snp);
    init_snp(snp, snp0, nsub, 1.0, 0.0, 0, 0);
    GENE1->first = snp;
    GENE1->last  = snp;
    GENE1->n_snps++;
  }
  if (n1) {
    snp = (SNP_STR *) malloc(sizeof(SNP_STR));
    CHECK_MEM(snp);
    init_snp(snp, snp1, nsub, 1.0, 0.0, 0, 0);
    if (GENE1->first == NULL) {
      GENE1->first = snp;
    } else {
      GENE1->last->next = snp;
    }
    GENE1->last = snp;
    GENE1->n_snps++;
  }
  if (n2) {
    snp = (SNP_STR *) malloc(sizeof(SNP_STR));
    CHECK_MEM(snp);
    init_snp(snp, snp2, nsub, 1.0, 0.0, 0, 0);
    if (GENE1->first == NULL) {
      GENE1->first = snp;
    } else {
      GENE1->last->next = snp;
    }
    GENE1->last = snp;
    GENE1->n_snps++;
  }
  free(snp0);
  free(snp1);
  free(snp2);



  /* Loop over non collapsed snps */
  SNP = pGENE->first;
  for (i=0; i<nsnp; i++) {
    k = cat0[i];
    if (k < 3) continue;
    snp = (SNP_STR *) malloc(sizeof(SNP_STR));
    CHECK_MEM(snp);
    init_snp(snp, SNP->G, nsub, SNP->weight, SNP->maf, 0, SNP->location);
    GENE1->last->next = snp;
    GENE1->last       = snp;
    GENE1->n_snps++;
    SNP = SNP->next;
  }

  info->gstar_invcor = dblMatrix_alloc(nsnp, nsnp);
  invcor = dblMatrix_alloc(ncat, ncat);
  cov = dblMatrix_alloc(ncat, ncat);

  i = gene_cov_invcor(GENE1, nsub, cov, useControls, controls, info->n_control, 1, invcor, &retInvFlag, NULL, info, 0);

  matrix_free((void **) cov, ncat);
  free_gene(GENE1, 1);
  if ((i) || (retInvFlag)) {
    info->errorFlag = i;
    matrix_free((void **) info->gstar_invcor, ncat);
    info->gstar_invcor = NULL;
    if (info->print > 1) printf_s("***ERROR***: G* correlation matrix for CMC test is singular. CMC test will not be computed.\n", NULL);
  } else {  
    /* Copy elements from invcor and info->geno_invcor to info->gstar_invcor */
    cor = info->gstar_invcor;
    freq012[0] = n0;
    freq012[1] = n1;
    freq012[2] = n2;
    for (i=0; i<nsnp; i++) {
      cati = cat0[i];
      for (j=i; j<nsnp; j++) {
        catj = cat0[j];
        if (cati != catj) {
          cor[i][j] = mat[i][j];
        } else if (cati > 2) {
          cor[i][j] = mat[i][j];
        } else if (freq012[cati] > 1) {
          cor[i][j] = invcor[cat[i]][cat[j]];
        } else {
          cor[i][j] = mat[i][j];
        }
        if (i != j) cor[j][i] = cor[i][j];
      }
    }  
  }

  free(cat);
  free(cat0);
  matrix_free((void **) invcor, ncat); 

} /* END: gstar_invcor */

/* Ad-hoc fix for when correlation matrix is singular */
static int fixInvCor(info, GENE, snpFreqFlag, controls, delta2Flag)
INFO *info;
GENE_STR *GENE;
int snpFreqFlag;
binary *controls;
int delta2Flag;
{
  int i, j, count, nsnp=GENE->n_snps, maxj, rm, nsub=info->nsubs;
  double **cor=info->geno_cor, maxcor, temp, *delta=info->delta, **cov=NULL, *sdG=NULL, vs=info->vs;
  binary *remove=NULL, retInvFlag, finalFlag;
  GENE_STR *GENE1=NULL;

  info->delta2_n = 0;

  if (cor == NULL) {
    if (info->print > 3) printf_s("***WARNING***: Input correlation matrix to fixInvCor is NULL\n", NULL);
    return(1);
  }

  remove = (binary *) malloc(nsnp*sizeof(binary));
  CHECK_MEM(remove);
  for (i=0; i<nsnp; i++) remove[i] = 0;

  /* Determine which snps to remove */
  count = nsnp;
  for (i=0; i<nsnp; i++) {
    if (!remove[i]) {
      maxcor = -9999.0;
      maxj = -1;
      for (j=0; j<nsnp; j++) {
        if ((i != j) && (!remove[j])) {
          temp = fabs(cor[i][j]);
          if (temp > maxcor) {
            maxcor = temp;
            maxj   = j;
          }
        }
      }
      if ((maxj > -1) && (maxcor > MAXCORRSNP)) {
        if (fabs(delta[i]) < fabs(delta[maxj])) {
          rm = i;
        } else {
          rm = maxj;
        }
        remove[rm] = 1;
        count--;
      }
    }
  }

  
  if ((count > 0) && (count < nsnp)) {
    info->delta2_n = count;

    /* Create a new gene structure with snps removed */
    GENE1 = (GENE_STR *) malloc(sizeof(GENE_STR));
    CHECK_MEM(GENE1);
    copyGeneStr(GENE, remove, nsub, GENE1, snpFreqFlag);
    if (info->GENE2) {
      free_gene(info->GENE2, 1);
      info->GENE2 = NULL;
    }
    info->GENE2 = GENE1;

    free(remove);
    remove = NULL;

    /* Allocate memory for cov and inverse corr */
    cov = dblMatrix_alloc(count, count);
    if (!info->geno_invcor2) info->geno_invcor2 = dblMatrix_alloc(count, count);

    /* Compute covariance matrix, inverse cor */
    i = gene_cov_invcor(GENE1, nsub, cov, info->useControlsForV, controls, info->n_control,\
                        1, info->geno_invcor2, &retInvFlag, NULL, info, snpFreqFlag);
    if ((i) || (retInvFlag)) {
      if (info->print > 3) printf_s("***ERROR***: Reduced inverse correlation matrix computation failed in fixInvCor\n", NULL);
      matrix_free((void **) cov, count);
      info->delta2_n = 0;
      matrix_free((void **) info->geno_invcor2, count);
      info->geno_invcor2 = NULL;
      free_gene(GENE1, 1);
      info->GENE2 = NULL;
      return(1);
    }

    if (!info->geno_sdG2) {
      info->geno_sdG2 = (double *) malloc(count*sizeof(double));
      CHECK_MEM(info->geno_sdG2);
    }
    sdG = info->geno_sdG2;
    for (i=0; i<count; i++) sdG[i] = sqrt(cov[i][i] + vs);
    matrix_free((void **) cov, count);
    cov = NULL;
    if (delta2Flag) {
      if (!info->delta2) {
        info->delta2 = (double *) malloc(count*sizeof(double));
        CHECK_MEM(info->delta2);
      }
      compute_delta(info->Y_norm, GENE1, info->delta2, sdG, info->weight_type);
    }
    finalFlag = 1;
  } else {
    finalFlag = 0;
  }

  if (cov != NULL) matrix_free((void **) cov, count);
  if (remove != NULL) free(remove);

  if (info->print > 3) {
    if (finalFlag) {
      printf_s("***NOTE***: Reduced inverse correlation matrix computation succeeded\n", NULL);
    } else {
      printf_s("***ERROR***: Reduced inverse correlation matrix computation failed. Some tests will not be computed.\n", NULL);
    }
  }

  return(0);

} /* END: fixInvCor */

/* Function to compute V, sigma, Y for the test stats for a gene */
int tests_init(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int nsnp=GENE->n_snps, nsub=info->nsubs, *pi1, i, j;
  TEST_STR **tests=info->tests, *tst;
  double *p, **V, vs=info->vs, *p2, temp;
  binary retInvFlag, invFlag=info->geno_invcor_flag;
  size_t memsize;

  info->errorFlag = 0;

  /* Set summary statistics to 0 */
  for (i=0; i<info->n_test; i++) {
    tst = tests[i];
    tst->nperm = 0;
    tst->cont = 0;
    tst->pvalue = MISS_PVALUE;
    if (!tst->len) {
      *tst->n_gt_obs = 0;
    } else {
      for (j=0, pi1=tst->n_gt_obs; j<tst->len; j++, pi1++) *pi1 = 0;
    }
  }

  /* Compute the covariance matrix and inverse correlation matrix */
  info->geno_cov = dblMatrix_alloc(nsnp, nsnp);
  if (invFlag) {
    info->geno_invcor = dblMatrix_alloc(nsnp, nsnp);
    info->geno_cor = dblMatrix_alloc(nsnp, nsnp);
   }
  i = gene_cov_invcor(GENE, nsub, info->geno_cov, info->useControlsForV, info->controls, info->n_control,\
                      invFlag, info->geno_invcor, &retInvFlag, info->geno_cor, info, 1);
  if (i) {
    /* Inverse correlation matrix would not be computed if cov failed */
    if (info->print > 1) printf_i("***ERROR***: In computing the covariance matrix. Error flag = %d\n", i);
    info->errorFlag = i;
    matrix_free((void **) info->geno_cov, nsnp);
    info->geno_cov = NULL;
    if (invFlag) {
      matrix_free((void **) info->geno_invcor, nsnp);
      info->geno_invcor = NULL;
    }
    return(i);
  }

  /* Compute the sd of geno_cov */
  info->geno_sdG = (double *) malloc(nsnp*sizeof(double));
  CHECK_MEM(info->geno_sdG);
  info->geno_varG = (double *) malloc(nsnp*sizeof(double));
  CHECK_MEM(info->geno_varG);
  V = info->geno_cov;
  for (i=0, p=info->geno_sdG, p2=info->geno_varG; i<nsnp; i++, p++, p2++) {
    temp = V[i][i] + vs;
    *p   = sqrt(temp);
    *p2  = temp;
  } 

  /* Compute the deltas */
  info->delta = (double *) malloc(nsnp*sizeof(double));
  CHECK_MEM(info->delta);
  compute_delta(info->Y_norm, GENE, info->delta, info->geno_sdG, info->weight_type);
  if ((invFlag) && (retInvFlag)) {
    info->errorFlag = retInvFlag;
    if (info->print > 1) printf_s("***ERROR***: Correlation matrix is singular\n", NULL);
    matrix_free((void **) info->geno_invcor, nsnp);
    info->geno_invcor = NULL;

    /* Try the ad-hoc fix */
    i = fixInvCor(info, GENE, 1, info->controls, 1);
  }

  /* Find SNPs with G=2 and compute the sum for Gik = 1 */
  /* if ((info->test_flag[TEST_SRT]) || (info->test_flag[TEST_SRTPOS])) find_G_equal2(GENE, nsub); */
 
  /* For the CMC test */
  if (info->test_flag[TEST_CMC]) {
    gstar_invcor(info, GENE);
  }
 
  /* For a user-defined R function */
  if (info->test_flag[TEST_R1]) {
    memsize = nsnp*sizeof(double);
    info->user_delta = (double *) malloc(memsize);
    CHECK_MEM(info->user_delta);
    memsize = nsnp*memsize;
    info->user_cov = (double *) malloc(memsize);
    CHECK_MEM(info->user_cov);
    info->user_cor = (double *) malloc(memsize);
    CHECK_MEM(info->user_cor);
    info->user_invcor = (double *) malloc(memsize);
    CHECK_MEM(info->user_invcor);
  }

  return(0);

} /* END: tests_init */

/***************************************/
/********* Test statistics *************/
/***************************************/

/* Sum test */
static double test_sum(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, j, nsnp=GENE->n_snps;
  double **V=info->geno_cov, test, num, den, den2, *p, *p2, *p3;

  num  = 0.0;
  den  = 0.0;
  den2 = 0.0;
  for (i=0, p=info->delta, p2=info->geno_sdG, p3=info->geno_varG; i<nsnp; i++, p++, p2++, p3++) {
    num += *p2 * *p;
    den += *p3;
    for (j=i+1; j<nsnp; j++) den2 += V[i][j];    
  }
  test = (2.0/nsnp)*num*num/(den + den2 + den2);

  return(test);

} /* END: test_sum */

/* Hotelling's Test */
static double test_hotelling(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, j, nsnp;
  double test, sumii, sumij, **sig, *delta, di, *sdG, sdi;

  if (info->geno_invcor) {
    sig   = info->geno_invcor;
    delta = info->delta;
    nsnp  = GENE->n_snps; 
    sdG   = info->geno_sdG;
  } else {
    sig   = info->geno_invcor2;
    delta = info->delta2;
    nsnp  = info->delta2_n; 
    sdG   = info->geno_sdG2;
  }

  sumii = 0.0;
  sumij = 0.0;
  for (i=0; i<nsnp; i++) {
    di  = delta[i];
    sdi = sdG[i];
    sumii += sdi*sdi*di*di;

    for (j=i+1; j<nsnp; j++) {
      sumij += sig[i][j]*di*delta[j];
    }
  }
  test = sumii + 2*sumij;
  /*if (info->CC_flag) test = test*info->test_cc_const;*/

  return(test);

} /* END: test_hotelling */

/* Hotelling's Test assuming independence */
static double test_hotIndep(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, nsnp=GENE->n_snps;
  double test, sumii=0.0, di, *pd;

  for (i=0, pd=info->delta; i<nsnp; i++, pd++) {
    di = *pd;
    sumii += di*di;
  }
  test = sumii;
  /*if (info->CC_flag) test = test*info->test_cc_const;*/

  return(test);

} /* END: test_hotIndep */

/* Stouffer's test */
static double test_stouffer(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, nsnp=GENE->n_snps;
  double test, sum=0.0, *p;

  for (i=0, p=info->delta; i<nsnp; i++, p++) sum += *p;
  test = sum*sum/nsnp;
  
  return(test);

} /* END: test_stouffer */

/* Fisher's test */
static double test_fisher(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, nsnp=GENE->n_snps;
  double test, sum=0.0, *delta=info->delta, temp;

  for (i=0; i<nsnp; i++) {
    temp = 2.0 - 2*normal_cdf(fabs(delta[i]), 0.0, 1.0);
    sum += log(temp);
  }
  test = -sum/nsnp;
  return(test);

} /* END: test_fisher */

/* Similarity Regression Test */
static double test_srt(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, nsnp=GENE->n_snps;
  double sum1, *p1, *p2;

  sum1 = 0.0;
  for (i=0, p1=info->delta, p2=info->geno_varG; i<nsnp; i++, p1++, p2++) sum1 += *p2 * *p1 * *p1;
    
  /*if (GENE->G_equal_2) sum1 += srt_sum2(info, GENE, info->Y_norm);*/
  
  return(sum1);

} /* END: test_srt */

/* Similarity Regression Test using positive deltas */
static double test_srtPos(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, nsnp=GENE->n_snps;
  double sum1, *p1, *p2, delta;

  sum1 = 0.0;
  for (i=0, p1=info->delta, p2=info->geno_varG; i<nsnp; i++, p1++, p2++) {
    delta = *p1;
    if (delta > 0.0) sum1 += *p2 *delta*delta;
  }
  
  /*if (GENE->G_equal_2) sum1 += srt_sum2(info, GENE, info->Y_norm);*/
  
  return(sum1);

} /* END: test_srt */

/* Data Adaptive Sum Test */
static double test_das(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, j, nsnp=GENE->n_snps;
  double test, sumii=0.0, *delta=info->delta, di, dj, temp, *sdG=info->geno_sdG, sdi;
  double wi, wj, wwi, wwj, sum1=0.0, sum2=0.0, phi1, phi2, vii, wd, wwd; 

  phi1 = info->qnorm_alpha;
  phi2 = -phi1;

  for (i=0; i<nsnp; i++) {
    di  = delta[i];
    sdi = sdG[i];
    vii = sdi*sdi; 
    sumii += vii*di*di;
    wi  = (di < phi1) ? -1.0 : 1.0;
    wwi = (di > phi2) ? -1.0 : 1.0;
    wd  = wi*di;
    wwd = wwi*di;
    
    for (j=i+1; j<nsnp; j++) {
      dj = delta[j];
      wj = (dj < phi1) ? -1.0 : 1.0;
      wwj = (dj > phi2) ? -1.0 : 1.0;      
      temp = sdi*sdG[j];
      sum1 += temp*wd*wj*dj;
      sum2 += temp*wwd*wwj*dj;
    }
  }
  temp = MAX(sum1, sum2);
  test = sumii + 2.0*temp;
  
  return(test);

}  /* END: test_das */

/* CMC */
static double test_cmc(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, j, nsnp;
  double **S=info->gstar_invcor, *delta, di, sum1, sum2, test;
  
  if (info->geno_invcor) {
    delta = info->delta;
    nsnp  = GENE->n_snps; 
  } else {
    delta = info->delta2;
    nsnp  = info->delta2_n; 
  }

  sum1 = 0.0;
  sum2 = 0.0;
  for (i=0; i<nsnp; i++) {
    di = delta[i];
    sum1 += S[i][i]*di*di;
    for (j=i+1; j<nsnp; j++) {
      sum2 += S[i][j]*di*delta[j];
    }
  }
  test = sum1 + sum2 + sum2;
  /*if (info->CC_flag) test *= info->test_cc_const;*/

  return(test);
}

/* BM */
static void test_bm(info, GENE, ret)
INFO *info;
GENE_STR *GENE;
double *ret;
{
  int i, j, n=info->bm_n, nsnp, ii;
  double di, dj, dci, sum1, sum2, *delta, h, cval, **S;  

  if (info->geno_invcor) {
    S     = info->geno_invcor;
    delta = info->delta;
    nsnp  = GENE->n_snps; 
  } else {
    S     = info->geno_invcor2;
    delta = info->delta2;
    nsnp  = info->delta2_n; 
  }

  if (n > 1) {
    h    = 4.0/(n - 1.0);
    cval = 0.0;
  } else {
    h    = 0.0;
    cval = 2.0;
  }

  /* Loop over each value of c */
  for (ii=0; ii<n; ii++) {
    sum1 = 0.0;
    sum2 = 0.0;
    for (i=0; i<nsnp; i++) {
      di = delta[i];
      if (di > 0.0) {
        dci = MAX(cval, di);
      } else {
        dci = MIN(-cval, di);
      }
      sum1 += dci*di;
      for (j=0; j<nsnp; j++) {
        if (j != i) {
          dj = delta[j];
          sum2 += S[i][j]*dci*dj;
        }
      }
    }
    ret[ii] = sum1 + sum2;
    cval += h;
  }
  
} /* END: test_bm */

/* Stouffers Positive Z */
static double test_stPosZ(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, nsnp=GENE->n_snps;
  double sum=0.0, test, *p;

  
  for (i=0, p=info->delta; i<nsnp; i++, p++) {
    if (*p > 0.0) sum += *p;
  }
  test = sum*sum/nsnp;

  return(test);
}

/* Threshold */
static double test_threshold(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, j, nsnp;
  double phi=info->threshold_phi, sum1=0.0, sum2=0.0, **S, *delta, di, dj, test;

  if (info->geno_invcor) {
    S     = info->geno_invcor;
    delta = info->delta;
    nsnp  = GENE->n_snps; 
  } else {
    S     = info->geno_invcor2;
    delta = info->delta2;
    nsnp  = info->delta2_n; 
  }

  for (i=0; i<nsnp; i++) {
    di = fabs(delta[i]);
    if (di > phi) {
      sum1 += di*di;
      for (j=i+1; j<nsnp; j++) {
        dj = fabs(delta[j]);
        if (dj > phi) sum2 += di*dj*fabs(S[i][j]);
      }
    }
  }
  test = (sum1 + sum2 + sum2)/nsnp;

  return(test);
}

/* ROVER */
static double test_rover(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, nsnp=GENE->n_snps;
  double w, di, test=0, t=info->rover_t, *p;

  for (i=0, p=info->delta; i<nsnp; i++, p++) {
    di = *p;
    di = di*di;
    w  = sqrt(1.0 - exp(-t*di));
    test += w*w*di;
  }

  return(test);
}

/* ROVER-V */
static double test_roverV(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, nsnp=GENE->n_snps;
  double w, *p, di, test=0, *p2, t=info->rover_t;

  for (i=0, p=info->delta, p2=info->geno_sdG; i<nsnp; i++, p++, p2++) {
    di = *p;
    di = di*di;
    w  = *p2 * sqrt(1.0 - exp(-t*di));
    test += w*w*di;
  }

  return(test);
}

/* Function to compute a single test */
static void compute_1test(info, GENE, which, ret)
INFO *info;
GENE_STR *GENE;
int which;
double *ret;
{

  switch (which) {
    case TEST_SUM:
      *ret = test_sum(info, GENE);
      break;
    case TEST_HOTELLING:
      *ret = test_hotelling(info, GENE);
      break;
    case TEST_STOUFFER:
      *ret = test_stouffer(info, GENE);
      break;
    case TEST_FISHER:
      *ret = test_fisher(info, GENE);
      break;
    case TEST_CMC:
      *ret = test_cmc(info, GENE);
      break;
    case TEST_SRT:
      *ret = test_srt(info, GENE);
      break;
    case TEST_SRTPOS:
      *ret = test_srtPos(info, GENE);
      break;
    case TEST_DAS:
      *ret = test_das(info, GENE);
      break;
    case TEST_BM:
      test_bm(info, GENE, ret);
      break;
    case TEST_STPOSZ:
      *ret = test_stPosZ(info, GENE);
      break;
    case TEST_THRESHOLD:
      *ret = test_threshold(info, GENE);
      break;
    case TEST_ROVER:
      *ret = test_rover(info, GENE);
      break;
    case TEST_ROVERV:
      *ret = test_roverV(info, GENE);
      break;
    case TEST_HOTINDEP:
      *ret = test_hotIndep(info, GENE);
      break;
    case TEST_C1:
      *ret = c1(info->delta, info->geno_cov, info->geno_cor, info->geno_invcor, GENE->n_snps);
      break;
    case TEST_C2:
      *ret = c2(info->delta, info->geno_cov, info->geno_cor, info->geno_invcor, GENE->n_snps);
      break;
    case TEST_C3:
      *ret = c3(info->delta, info->geno_cov, info->geno_cor, info->geno_invcor, GENE->n_snps);
      break;
    case TEST_C4:
      *ret = c4(info->delta, info->geno_cov, info->geno_cor, info->geno_invcor, GENE->n_snps);
      break;
    case TEST_C5:
      *ret = c5(info->delta, info->geno_cov, info->geno_cor, info->geno_invcor, GENE->n_snps);
      break;
    default:
      #if RFLAG
      *ret = user_r_main(info, GENE, which); 
      #else 
      *ret = 0.0;
      #endif
      break;
  } 

} /* compute_1test */

/* Function to compute the test statistics given Y and delta in info */
void compute_tests2(info, GENE, permFlag)
INFO *info;
GENE_STR *GENE;
binary permFlag;  /* 0 = observed, 1 = permutation */
{
  int i;
  TEST_STR **tests=info->tests, *tst;
  double *p;

  for (i=0; i<info->n_test; i++) {
    tst = tests[i];
    if (!tst->cont) {
      continue;
    } else {
      p = (permFlag) ? tst->test : tst->obs;
      compute_1test(info, GENE, tst->which, p);
    }
  }

} /* compute_tests2 */

/* Function to compute the adaptive stopping criteria */
static int adapt_continue(pval, nperm, tol)
double pval, tol;
int nperm;
{
  double temp;

  if (pval < NUM_ZERO) return(1);
  temp = 3.92*sqrt(pval*(1.0-pval)/nperm)/pval;
  if (temp < tol) {
    return(0);
  } else {
    return(1);
  }
  
} /* END: adapt_continue */

/* Function to compute snp variances using the controls for a permutation */
static void compute_sdG_CC0(controls, nsub, GENE, sdG, sdG0, ncontrol, vs, varG)
binary *controls;
int nsub;
GENE_STR *GENE;
double *sdG;
double *sdG0; /* Observed values */
int ncontrol;
double vs;
double *varG;  /* Variance */
{
  SNP_STR *SNP;
  int i, j, n, *pi1, m, m1=0, m2=0, mmiss;
  double sum, sum2, mean, var, temp;

  SNP = GENE->first;
  for (i=0; i<GENE->n_snps; i++) {
    m1 = m2 = mmiss = 0;

    n = SNP->n1;
    if (n) {
      for (j=0, pi1=SNP->id1; j<n; j++, pi1++) {
        if (controls[*pi1]) m1++;
      }
    } 
    n = SNP->n2;
    if (n) {
      for (j=0, pi1=SNP->id2; j<n; j++, pi1++) {
        if (controls[*pi1]) m2++;
      }
    }
    m = m1 + m2;
    sum = 2.0*m2 + m1;
    sum2 = 4.0*m2 + m1;

    /* Get the number of zero genotypes */
    if (m < ncontrol) {
      n = SNP->nmiss;
      if (n) {
        for (j=0, pi1=SNP->idmiss; j<n; j++, pi1++) {
          if (controls[*pi1]) mmiss++;
        }
      }
      m += ncontrol - m - mmiss;
    }
/*
printf("m=%d, m1=%d, m2=%d, sum=%g, sum2=%g, nmiss=%d, ncontrols=%d\n", m, m1, m2, sum, sum2, mmiss, ncontrol);
*/
    if (m > 1) {
      mean = sum/m;
      var = (sum2 - m*mean*mean)/(m - 1.0);
      if (var > ALMOST_ZERO) {
        temp    = var + vs;
        sdG[i]  = sqrt(temp);
        if (varG) varG[i] = temp;
      } else {
        temp    = sdG0[i];
        sdG[i]  = temp;
        if (varG) varG[i] = temp*temp; 
      }
    } else {
      temp    = sdG0[i];
      sdG[i]  = temp;
      if (varG) varG[i] = temp*temp; 
    }   

    SNP = SNP->next;
  }

} /* END: compute_sdG_CC0 */

/* Function to compute V, sigma, Y for the test stats for a gene */
static int compute_perm_cov(info, GENE, controls)
INFO *info;
GENE_STR *GENE;
binary *controls;
{
  int nsnp=GENE->n_snps, nsub=info->nsubs, i, ret=0;
  double *p, **V, vs=info->vs, temp, *p2;
  binary retInvFlag, invFlag=info->geno_invcor_flag;

  /* Compute the covariance matrix and inverse correlation matrix */
  if (!info->geno_cov) info->geno_cov = dblMatrix_alloc(nsnp, nsnp);
  if (invFlag) {
    if (!info->geno_invcor) info->geno_invcor = dblMatrix_alloc(nsnp, nsnp);
    if (!info->geno_cor) info->geno_cor = dblMatrix_alloc(nsnp, nsnp);
   }
  ret = gene_cov_invcor(GENE, nsub, info->geno_cov, info->useControlsForV, controls, info->n_control,\
                      invFlag, info->geno_invcor, &retInvFlag, info->geno_cor, info, 1);
  if (ret) {
    /* Inverse correlation matrix would not be computed if cov failed */
    if (info->print > 1) printf_i("***ERROR***: In computing the covariance matrix. Error flag = %d\n", i);
    /*
    matrix_free((void **) info->geno_cov, nsnp);
    info->geno_cov = NULL;
    if (invFlag) {
      matrix_free((void **) info->geno_invcor, nsnp);
      info->geno_invcor = NULL;
    }
    */
    return(ret);
  } 
/*
print_dMat(info->geno_cov, GENE->n_snps, GENE->n_snps, "cov");
*/

  if ((invFlag) && (retInvFlag)) {
    if (info->print > 1) printf_s("***ERROR***: Correlation matrix is singular\n", NULL);
    /*
    matrix_free((void **) info->geno_invcor, nsnp);
    info->geno_invcor = NULL;
    */

    /* Try the ad-hoc fix. Do not update delta2 here, it will be done in permute_tests */
    ret = fixInvCor(info, GENE, 1, controls, 0);
  } else {
    /* Compute the sd of geno_cov */
    if (!info->geno_sdG) {
      info->geno_sdG = (double *) malloc(nsnp*sizeof(double));
      CHECK_MEM(info->geno_sdG);
    }
    if (!info->geno_varG) {
      info->geno_varG = (double *) malloc(nsnp*sizeof(double));
      CHECK_MEM(info->geno_varG);
    }

    V = info->geno_cov;
    for (i=0, p=info->geno_sdG, p2=info->geno_varG; i<nsnp; i++, p++, p2++) {
      temp = V[i][i] + vs;
      *p = sqrt(temp);
      *p2 = temp;
    }
  }
 
  return(ret);

} /* END: compute_perm_cov */

/* Function to decide to update the correlation matrix */
static int updateCorr(obs, perm, cutoff)
double obs, perm, cutoff;
{
  if (perm > ALMOST_ZERO) {
    if (fabs(obs - perm)/perm < cutoff) return(1); 
  } else if (obs > ALMOST_ZERO) {
    if (fabs(obs - perm)/obs < cutoff) return(1); 
  } else {
    if (0.0 < cutoff) return(1);
  }

  return(0);

} /* END: updateCorr */

/* Function to return binary vector for controls */
static void getCutoffVec(vec, n, cutoff, bvec)
double *vec, cutoff;
int n;
binary *bvec;
{
  int j;
  double *pd;
  binary *pb;

  for (j=0, pb=bvec, pd=vec; j<n; j++, pb++, pd++) {
    if (*pd < cutoff) {
      *pb = 1;
    } else {
      *pb = 0;
    }
  } 

} /* END: getCutoffVec */

/* Function to get response vector for each perm */
static double * getY_perm(Y, nsub, permMat, row)
double *Y;
int nsub, row;
double **permMat;
{

  if (!permMat) {
    permute_dVec(Y, nsub, NULL);
  } else {
    Y = permMat[row];
  }

  return(Y);

} /* END: getY_perm */

/* Function to permute Y and calculate tests for a gene */
void permute_tests(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, j, ii, nsub=info->nsubs, nsnp=GENE->n_snps, *pi, nperm=info->nperm, permCount, newCorrCount=0, totalCount=0;
  int ntest=info->n_test, bm_n=info->bm_n, ncont, k, ncontrol=info->n_control; 
  double *delta=info->delta, *delta0, *Y=info->Y_norm, *Y0, *p, *p1, *p2;
  double tol=info->adapt_tol, *tmp=NULL, temp, *sdG=info->geno_sdG, *delta2=info->delta2, *sdG2=info->geno_sdG2;
  binary bmFlag=info->test_flag[TEST_BM],  adaptFlag=info->adapt_flag, delta2Flag=0, sdG_flag, *controls=NULL;
  TEST_STR **tests=info->tests, *tst;
  GENE_STR *GENE2=info->GENE2;
  double *sdG_0, *sdG2_0, vs=info->vs, hot_obs, update_cor=info->update_cor, *varG=info->geno_varG;
  binary print=info->print, update_cor_flag=1;
  double **permMat=info->storePermMat;

  /* NOTE: Can be improved by checking if a test that requires the covariance matrix stops early */
  if ((info->geno_cov_flag == 0) && (info->geno_cor_flag == 0) && (info->geno_invcor_flag == 0)) update_cor_flag = 0;
  if (info->update_cor < NUM_ZERO) update_cor_flag = 0;

  if (info->delta2_n) delta2Flag = 1;
  if ((info->CC_flag) && (info->useControlsForV)) {
    sdG_flag = 1;
  } else {
    sdG_flag = 0;
  }

  Y0 = (double *) malloc(nsub*sizeof(double));
  CHECK_MEM(Y0);
  delta0 = (double *) malloc(nsnp*sizeof(double));
  CHECK_MEM(delta0);
  if (sdG_flag) {
    sdG_0 = (double *) malloc(nsnp*sizeof(double));
    CHECK_MEM(sdG_0);
    if (delta2Flag) {
      sdG2_0 = (double *) malloc(info->delta2_n*sizeof(double));
      CHECK_MEM(sdG2_0);
    }
    controls = (binary *) malloc(nsub*sizeof(binary));
    CHECK_MEM(controls);
  }

  if (bmFlag) {
    tmp = (double *) malloc(bm_n*sizeof(double));
    CHECK_MEM(tmp);
  }

  /* Save original Y and delta */
  for (i=0, p1=Y0, p2=Y; i<nsub; i++, p1++, p2++) *p1 = *p2;
  for (i=0, p1=delta0, p2=delta; i<nsnp; i++, p1++, p2++) *p1 = *p2;
  if (sdG_flag) {
    for (i=0, p1=sdG_0, p2=sdG; i<nsnp; i++, p1++, p2++) *p1 = *p2;
    if (delta2Flag) {
      for (i=0, p1=sdG2_0, p2=sdG2; i<info->delta2_n; i++, p1++, p2++) *p1 = *p2;
    } 
  }

  /* Compute the observed Hotelling Test */
  if (update_cor_flag) hot_obs = test_hotelling(info, GENE);

  for (ii=0; ii<info->adapt_maxiter; ii++) {
    permCount = 0;
    for (i=0; i<nperm; i++) {

      /* Permute Y  */
      Y = getY_perm(Y, nsub, permMat, i);
      /*permute_dVec(Y, nsub, NULL);*/

      /* Determine if we need to update the correlation matrix */
      if (sdG_flag) {
        /* Get vector of controls for this permutation */
        getCutoffVec(Y, nsub, 0.5, controls);

        /* Compute the standard deviations sdG. These will be used in the deltas */

        compute_sdG_CC0(controls, nsub, GENE, sdG, sdG_0, ncontrol, vs, varG);
        if (delta2Flag) compute_sdG_CC0(controls, nsub, GENE2, sdG2, sdG2_0, ncontrol, vs, NULL);

        /*
        compute_sdG_CC(controls, nsub, GENE, sdG, sdG_0, vs);
        if (delta2Flag) compute_sdG_CC(controls, nsub, GENE2, sdG2, sdG2_0, vs);
        */

        /* Compute delta */
        compute_delta(Y, GENE, delta, sdG, info->weight_type);
        if (delta2Flag) compute_delta(Y, GENE2, delta2, sdG2, info->weight_type);

        /* Code for updating the genotype covariance matrix */
        if (update_cor_flag) {
          /* Compute the Hotelling Test */
          temp = test_hotelling(info, GENE);

          if (updateCorr(hot_obs, temp, update_cor)) { 
            /* Compute new covariance, correlation, and inverse for each permutation */
            newCorrCount++;
            /*if (print > PRINT_NEWCORR) printf_s("New correlation matrix will be computed\n", NULL);*/
            j = compute_perm_cov(info, GENE, controls);
            if (j) continue;
          }
        }
      } else {
        /* Compute delta */
        compute_delta(Y, GENE, delta, sdG, info->weight_type);
        if (delta2Flag) compute_delta(Y, GENE2, delta2, sdG2, info->weight_type);
      }

      /* Compute the test statistics */
      compute_tests2(info, GENE, 1);
      permCount++;

      /* Update summary statistics */
      for(j=0; j<ntest; j++) {
        tst = tests[j];
        if (tst->cont) {
          if (!tst->len) {
            if (*tst->test > *tst->obs) (*tst->n_gt_obs)++;
          } else {
            for (k=0, p1=tst->obs, p2=tst->test, pi=tst->n_gt_obs; k<tst->len; k++, p1++, p2++, pi++) {
              if (*p2 > *p1) (*pi)++;
            }
          }
        }
      }
      /*
      if (perm_fpt) {
        fprintf(perm_fpt, "%s\t%d", GENE->name, i+1);
        for(j=0; j<ntest; j++) fprintf(perm_fpt, "\t%g", *(tests[j]->test));
        fprintf(perm_fpt, "\n");
        fflush(perm_fpt);
      }
      */

    } /* END: for (i=0; i<nperm; i++) */

    /* Update number of permutations, calculate p-values, and determine if the test should continue */
    totalCount += permCount;
    ncont = 0;
    for(j=0; j<ntest; j++) {
      tst = tests[j];
      if (tst->cont) {
        tst->nperm += permCount;
        if (!tst->len) {
          tst->pvalue = ((double) *tst->n_gt_obs)/tst->nperm;
        } else {
          temp = 1.0/tst->nperm;
          for (k=0, pi=tst->n_gt_obs, p=tmp; k<tst->len; k++, pi++, p++) *p = temp * *pi;
          tst->pvalue = min_dvec(tmp, tst->len);
        }
        if (adaptFlag) {
          k = adapt_continue(tst->pvalue, tst->nperm, tol);
          tst->cont = k;
          ncont += k; 
        }
      }
    } /* END: for(j=0; j<ntest; j++) */
   
    if (!ncont) break;

  } /* END: for (ii=0; ii<info->adapt_maxiter; ii++) */

  if (print > 3) printf_d("New correlation matrix computed in %g percent of the permutations\n", (100.0*newCorrCount)/totalCount);

  /* Reset Y and delta */
  for (i=0, p1=info->Y_norm, p2=Y0; i<nsub; i++, p1++, p2++) *p1 = *p2;
  free(Y0);

  for (i=0, p1=delta, p2=delta0; i<nsnp; i++, p1++, p2++) *p1 = *p2;
  free(delta0);
  if (sdG_flag) {
    for (i=0, p1=sdG_0, p2=sdG; i<nsnp; i++, p1++, p2++) *p2 = *p1;
    free(sdG_0);
    if (delta2Flag) {
      for (i=0, p1=sdG2_0, p2=sdG2; i<info->delta2_n; i++, p1++, p2++) *p2 = *p1;
      free(sdG2_0);
    }
  } 
  if (controls) free(controls);
  if (tmp) free(tmp);

} /* END: permute_tests */

/* Function to compute all the effect sizes for the snps in a gene */
void effect_size_all(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  int i, j, k, nsnp=GENE->n_snps, nsub=info->nsubs, ncov=info->n_use_covars, nr, nc;
  double **design, **covmat=info->covmat, *response, **prow, *pcol, *pr, **pd2, *pd, *beta, **xpx=NULL;
  SNP_STR *SNP=GENE->first;
  genoType *pg, gg;
  yType *y=info->CC, *py;
  binary CC_flag=info->CC_flag, simpleFlag=0;
  
  nc = ncov + 2; /* For intercept and SNP */
  response = (double *) malloc(nsub*sizeof(double));
  CHECK_MEM(response);
  design = dblMatrix_alloc(nsub, nc);
  beta = (double *) malloc(nc*sizeof(double));
  CHECK_MEM(beta);
  if (!ncov) {
    simpleFlag = 1;
    xpx = dblMatrix_alloc(nc, nc);
  }

  for (i=0; i<nsnp; i++) {
    nr  = 0;
    pr  = response;
    pd2 = design;

    /* Set up design matrix and response */
    for (j=0, pg=SNP->G, py=y, prow=covmat; j<nsub; j++, pg++, py++, prow++) {
      gg = *pg;
      if (gg != MISS_GENO) {
        /* Use this row */
        nr++;
        *pr++ = *py;
        pd    = *pd2;
        *pd++ = 1.0;
        *pd++ = (double) gg; 
        if (ncov) {    
          for (k=0, pcol=*prow; k<ncov; k++, pcol++) *pd++ = *pcol; 
        } 
        pd2++;
      }
    }

    if (simpleFlag) {
      xpx[0][0] = nr;
      xpx[0][1] = xpx[1][0] = SNP->n1 + 2*SNP->n2;
      xpx[1][1] = SNP->n1 + 4*SNP->n2; 
    }

    if (CC_flag) {
      j = beta_logReg(design, nr, nc, response, beta, xpx);
    } else {
      j = beta_normal(design, nr, nc, response, beta, xpx);
    }

    if (!j) {
      SNP->effect_size = exp(beta[1]);
    } else {
      SNP->effect_size = ERROR_EFFECT_SIZE;
    }
    SNP = SNP->next;
  }

  free(response);
  matrix_free((void **) design, nsub);
  free(beta);
  if (xpx) matrix_free((void **) xpx, nc);

} /* END: effect_size_all */



