/***************************************************************
                                                             
                    CRAVE                  
                                                              
 This file is distributed under the GNU General Public        
 License, Version 2. See the file COPYING.txt for more    
 details                                                      
                                                              
****************************************************************/

#include "./crave.h"

/*****************************************************************************
***************** Memory allocation*******************************************
******************************************************************************/

int ** intMatrix_alloc(nrow, ncol)
int nrow, ncol;
{
  int **mat, **ptr;
  int i;
  size_t size;

  mat = (int **) malloc(nrow*sizeof(int *));
  CHECK_MEM(mat);
  size = ncol*sizeof(int);
  for (i=0, ptr=mat; i<nrow; i++, ptr++) {
    *ptr = (int *) malloc(size);
    CHECK_MEM(*ptr);
  }
  return(mat);
}

double ** dblMatrix_alloc(nrow, ncol)
int nrow, ncol;
{
  double **mat, **ptr;
  int i;
  size_t size;

  mat = (double **) malloc(nrow*sizeof(double *));
  CHECK_MEM(mat);
  size = ncol*sizeof(double);
  for (i=0, ptr=mat; i<nrow; i++, ptr++) {
    *ptr = (double *) malloc(size);
    CHECK_MEM(*ptr);
  }
  return(mat);
}

binary ** binMatrix_alloc(nrow, ncol)
int nrow, ncol;
{
  binary **mat, **ptr;
  int i;
  size_t size;

  mat = (binary **) malloc(nrow*sizeof(binary *));
  CHECK_MEM(mat);
  size = ncol*sizeof(binary);
  for (i=0, ptr=mat; i<nrow; i++, ptr++) {
    *ptr = (binary *) malloc(size);
    CHECK_MEM(*ptr);
  }
  return(mat);
}

/* Function to allocate memory for a char vector */
char ** cVec_alloc(n, len)
int n, len;
{
  int i;
  char **ret;

  ret = (char **) malloc(n*sizeof(char *));
  CHECK_MEM(ret);
  for (i=0; i<n; i++) {
    ret[i] = (char *) malloc(len*sizeof(char));
    CHECK_MEM(ret[i]);
  }
  
  return(ret);

} /* END: cVec_alloc */

/* Function to allocate memory for a char vector */
char * c0Vec_alloc(n)
int n;

{
  char *ret;

  ret = (char *) malloc(n*sizeof(char));
  CHECK_MEM(ret);
  
  return(ret);

} /* END: c0Vec_alloc */

/* Function to allocate memory for a double vector */
double * dVec_alloc(n, initFlag, initVal)
int n, initFlag;
double initVal;
{
  int i;
  double *ret, *p;

  ret = (double *) malloc(n*sizeof(double));
  CHECK_MEM(ret);
  if (initFlag) {
    for (i=0, p=ret; i<n; i++, p++) *p = initVal;
  }
  
  return(ret);

} /* END: dVec_alloc */

/* Function to allocate memory for a integer vector */
int * iVec_alloc(n, initFlag, initVal)
int n, initFlag, initVal;
{
  int i, *ret, *p;

  ret = (int *) malloc(n*sizeof(int));
  CHECK_MEM(ret);
  if (initFlag) {
    for (i=0, p=ret; i<n; i++, p++) *p = initVal;
  }
  
  return(ret);

} /* END: iVec_alloc */

/* Function to free a matrix */
void matrix_free(x, n)
void **x;
int n;
{
  int i;
  for (i=0; i<n; i++) {
    if (x[i] != NULL) free(x[i]);
  }
  free(x);

} /* END: matrix_free */


/* Function to initialize a binary vector to a constant */
void bvecinit(vec, n, value)
binary *vec, value;
int n;
{
  binary *ptr = vec;
  while (n-- > 0) {
    *ptr++ = value;
  }
}

/* Function to copy a gene structure */
void copyGeneStr(GENE0, exclude, nsub, GENE1, snpFreqFlag)
GENE_STR *GENE0, *GENE1;
int nsub;
binary *exclude;
int snpFreqFlag; /* 0 or 1 to compute genotype frequncy counts */
{
  int nsnp=GENE0->n_snps, i, n;
  SNP_STR *SNP0, *snp1;
  binary *pb, firstFlag=0;

  SNP0 = GENE0->first;
  n = 0;
  for (i=0, pb=exclude; i<nsnp; i++, pb++) {
    if (!*pb) {
      GENE1->G_equal_2 = GENE0->G_equal_2;
      GENE1->next = NULL;
      GENE1->previous = NULL;
      GENE1->name = NULL;      
      n++;

      /* Allocate SNP structure */
      snp1 = (SNP_STR *) malloc(sizeof(SNP_STR));
      CHECK_MEM(snp1);
      init_snp(snp1, SNP0->G, nsub, SNP0->weight, SNP0->maf, snpFreqFlag, SNP0->location);
      snp1->G_equal_2 = SNP0->G_equal_2;
      if (!firstFlag) {
        GENE1->first = snp1; 
        GENE1->last  = snp1;
        firstFlag   = 1;
      } else {
        GENE1->last->next = snp1;
        GENE1->last       = snp1;
      }
    }
    SNP0 = SNP0->next;
  }
  GENE1->n_snps = n;
  
} /* END: copyGeneStr */

/*****************************************************************************
***************** String functions *******************************************
******************************************************************************/

/* Function to get the next token */
char * next_token(str, n, sep)
char *str, sep;
long n;
{
  int i;
  char *pc, c1;

  if (str == NULL) return(NULL);
  
  /* Loop till a sep character or end of string */
  for (i=0, pc=str; i<n; i++, pc++) {
    c1 = *pc;
    if (c1 == sep) {
      pc++;
      return(pc);
    } else if (c1 == END_OF_STR) {
      return(NULL);
    }
  }
  return(NULL);

} /* END: next_token */

/* Function to copy a token from a line read in to a field and adding '\0' */
int copy_token(ret, str, sep, n)
char *ret; /* Return string */
char *str; /* Must start at the correct position */
char sep;
long n;
{
  int i, flag=2;
  char *pc, *pret, c1;
  
  /* The end-of-line character should appear before the end-of-string character */
 
  /*n = MIN(n, strlen(str));*/
  for (i=0, pc=str, pret=ret; i<n; i++, pc++, pret++) {
    c1 = *pc;

    if ((c1 != sep) && (c1 != END_OF_LINE) && (c1 != CAR_RETURN) && (c1 != EOF)) {
      *pret = *pc;
      flag = 1;
    } else {
       *pret = END_OF_STR;
       flag--;
       break;
    }
  }

  return(flag);

} /* END: copy_token */

/* Function to count a char in a string */
int count_char(str, c1)
char *str, c1;
{
  int i, len, n;

  len = strlen(str);
  n = 0;
  for (i=0; i<len; i++) {
    if (str[i] == c1) n++;
  }

  return(n);

} /* END: count_char */

/* Function to count a char in a string */
int count_char_maxlen(str, c1, maxlen)
char *str, c1;
int *maxlen;
{
  int i, len, n, max, toklen;

  len = strlen(str);

  n = 0;
  max = -1;
  toklen = 0;
  for (i=0; i<len; i++) {
    toklen++;
    if (str[i] == c1) {
      n++;
      if (toklen > max) max = toklen;
      toklen = 0;
    }
    if (toklen > max) max = toklen;
  }
  if (!n) max = len;

  *maxlen = max;
  return(n);

} /* END: count_char */

/* Function to compare character match character vectors with unique elements in each vec */
int matchCharVecs(vec1, use1, n1, vec2, n2, ret_v1, ret_v2, v1_v2) 
char **vec1, **vec2;
int n1, n2;
binary *ret_v1, *ret_v2;
binary *use1; /* Vector of which elements in vec1 to use */
int *v1_v2; /* Length n1 vector containing position of v1 in v2 */
{
  int i, j, *p12, nmatch;
  binary *pb1, *pb2, *pu1;
  char **pc1, **pc2, *pchar;

  /* Initialize ret_v1 and ret_v2 to 0 */
  bvecinit(ret_v1, n1, 0);
  bvecinit(ret_v2, n2, 0);
  nmatch = 0;

  for (i=0, pc1=vec1, pb1=ret_v1, p12=v1_v2, pu1=use1; i<n1; i++, pc1++, pb1++, p12++, pu1++) {
    if (*pu1) {
      pchar = *pc1;
      for (j=0, pb2=ret_v2, pc2=vec2; j<n2; j++, pb2++, pc2++) {
        if (!*pb2) {
          if (!strcmp(pchar, *pc2)) {
            *pb2 = 1;
            *pb1 = 1;
            *p12 = j;
            nmatch++;
            break; 
          }
        }
      }
    } 
  }

  return(nmatch);

} /* END: matchCharVecs */

/* Function to parse a delimited string into a vector. Memory is allocated for the return vector */
char ** strIntoVec(str, sep, ret_n)
char *str, sep;
int *ret_n;
{
  char *token, **ret_vec;
  int i, n, maxlen;
  size_t size;

  /* Get the number of sub-strings */
  n = count_char_maxlen(str, sep, &maxlen) + 1;

  maxlen = maxlen + 1;

  *ret_n = n;

  ret_vec = (char **) malloc(n*sizeof(char *));
  CHECK_MEM(ret_vec);
  size = sizeof(char);
  token = str;

  for (i=0; i<n; i++) {
    ret_vec[i] = (char *) malloc(maxlen*size);
    CHECK_MEM(ret_vec[i]); 

    copy_token(ret_vec[i], token, sep, maxlen);

    token = next_token(token, maxlen, sep);
  } 

  return(ret_vec);

} /* END: strIntoVec */


/* Find the index of a string in a character vector. Strings must end with \0 */
int strIndexInVec(str, vec, n)
char *str;
char **vec;
int n;
{
  int i, ret=-1, j;
  char **pv;

  for (i=0, pv=vec; i<n; i++, pv++) {
    j = strcmp(str, *pv);
    if (!j) return(i);
  }

  return(ret);

} /* strIndexInVec */

/* Function to return the variable number starting at 1 */
int getVarNum(var, varVec, n, exitOnError)
char *var, **varVec;
int n, exitOnError;
{
  int ret;
  char *end;

  /* First determine if var is an integer */
  ret = strtol(var, &end, 10); 
  if ((ret > n) && (exitOnError)) {
    printf_s("ERROR with variable %s\n", var);
    exit(1);
  }
  if (ret > 0) return(ret);
  
  ret = strIndexInVec(var, varVec, n);
  if (ret > -1) {
    ret += 1;
  } else if (exitOnError) {
    printf_s("ERROR with variable %s\n", var);
    exit(1);
  }

  return(ret);

} /* END: getVarNum */

/* Find the index of a string in a character vector updating another vector */
int strIndexInVec2(str, vec, n, found)
char *str;
char **vec;
int n;
binary *found;
{
  int i, ret=-1, j;
  char **pv;
  binary *pb;

  for (i=0, pv=vec, pb=found; i<n; i++, pv++, pb++) {
    if (!*pb) {
      j = strcmp(str, *pv);
      if (!j) {
        *pb = 1;
        return(i);
      }
    }
  }

  return(ret);

} /* strIndexInVec2 */

/* Function to remove white space from a string */
char * rmWhiteSpace(str)
char *str;
{
  char *pc, c1;
  int len, i;

  if ((str == NULL) || (str == END_OF_STR)) return(str);

  pc = str;
  while (1) {
    c1 = *pc;
    if ((c1 == ' ') || (c1 == '\t')) {
      pc++;
    } else {
      break;
    }    
  }

  len = (int) strlen(pc);

  for (i=len-1; i>-1; i--) {
    c1 = pc[i];
    if ((c1 == END_OF_STR) || (c1 == END_OF_LINE) || (c1 == CAR_RETURN) || (c1 == ' ') || (c1 == '\t')) {
      pc[i] = END_OF_STR;
    } else {
      break;
    }

  }  

  return(pc);

} /* END: rmWhiteSpace */

/* Function to find a string in a character vector */
int findStrInVec(str, vec, n)
char *str;
char **vec;
int n;
{
  int i, i1, i2, mid, ret1, ret2, ret, j;
  
  if (str == NULL) return(-1);
  if (n < 4) {
    for (i=0; i<n; i++) {
      if (!strcmp(str, vec[i])) return(i);
    }
    return(-1);
  }
  
  /* Check the end points */
  i1   = 0;
  i2   = n-1;
  ret1 = strcmp(str, vec[i1]);
  if (ret1 == 0) {
    return(i1);
  } else if (ret1 < 0) {
    return(-1);
  }
  ret2 = strcmp(str, vec[i2]);
  if (ret2 == 0) {
    return(i2);
  } else if (ret2 > 0) {
    return(-1);
  }

  for (i=0; i<n; i++) {
    if (i2 - i1 < 4) {
      for (j=i1+1; j<=i2-1; j++) {
        if (!strcmp(str, vec[j])) return(j);
      }      
      return(-1);
    }

    mid = ceil(0.5*(i1 + i2));
    ret = strcmp(str, vec[mid]);
    if (!ret) {
      return(mid);
    } else if (ret > 0) {
      i1 = mid;
    } else {
      i2 = mid;
    }
  }

  return(-1);

} /* END: findStrInVec */

/* Function to return 1 if an id is to be removed */
int removeStr(str, vec, n, includeFlag)
char *str;
char **vec;
int n;
binary includeFlag;
{
  int ret;

  ret = findStrInVec(str, vec, n);
  if (includeFlag) {
    if (ret < 0) {
      return(1);
    } else {
      return(0);
    }
  } else {
    if (ret < 0) {
      return(0);
    } else {
      return(1);
    }
  }

} /* END: removeStr */

/* Function to extract a substring between 2 delimiters */
char * extractStr2(str, exStr, del1, del2)
char *str, *exStr, *del1, *del2;
{
  char *p, *token;

  p = strstr(str, exStr);
  token = strtok(p, del1);
  token = strtok(NULL, del2);
  return(token);

} /* END: extractStr2 */

/* Function to extract a substring between 2 1 char delimiters ex: reflink.geneName=10294;
   Second delimiter could also be a tab */
int getStr2delim(str, len, del1, del2, sep, ret_str)
char *str, *del1, *del2, *ret_str, sep;
int len;
{
  int i, j;
  char *begin, c, sep2=*del2;

  j = 0;
  len = MIN(len, strlen(str));
  begin = strstr(str, del1); 
  if (begin == NULL) return(1);
  begin++;
  for (i=0; i<len; i++) {
    c = begin[i];
    if ((c != sep2) && (c != sep)) {
      ret_str[j++] = c;
    } else {
      ret_str[j] = END_OF_STR;
      break;
    }
  }

  return(0);

} /* END: getStr2delim */

/* Function to get the directory of a file from a string
   ret must have sufficent memory allocated.
   The default for ret is "./", the current directory */
void getDirName(str, len, ret)
char *str, *ret;
int len;  /* Length of str */
{
  int i, j;
  binary flag=0;

  for (i=len-1; i>-1; i--) {
    if (str[i] == '/') {
      for (j=0; j<=i; j++) ret[j] = str[j];
      ret[i+1] = END_OF_STR;
      flag = 1;
      break;
    }
  }
  if (!flag) {
    ret[0] = '.';
    ret[1] = '/';
    ret[2] = END_OF_STR;
  }

} /* END: getDirName */

/* Function to get the integers from a string such as "1-7,13,20-25" */
int getIntsFromStr(str, intLen, n, retVec)
char *str;
int intLen; /* Maximum number of characters for one of the ints */
int n;  /* Length of retVec (1 + max number) */
binary *retVec;
{
  int i, strIndex, len, index, start, num, rangeFlag, im1;
  char *end, chr[intLen+1], c1;

  /* Initialize */
  for (i=0; i<n; i++) retVec[i] = 0;
  len = (int) strlen(str);
  
  strIndex = 0;
  start = 0;
  rangeFlag = 0;
  index = 0;
  c1 = 'e';
  im1 = intLen - 1;

  while(1) {
    if (strIndex <= len) c1 = str[strIndex];

    if ((strIndex > len) || (c1 == ',')) {

      if (!index) {
        printf_s("\nERROR: The list %s is not valid\n", str);
        exit(1);
      }

      /* Get number */
      chr[index] = END_OF_STR;
      num = (int) strtol(chr, &end, 10);
      if ((num < 1) || (num > n)) {
        printf_s("\nERROR: The list %s is not valid\n", str);
        exit(1);
      }

      if (rangeFlag) {
        if ((!start) || (num < start)) {
          printf_s("\nERROR: The list %s is not valid\n", str);
          exit(1);
        }
        for (i=start; i<=num; i++) retVec[i] = 1;
      } else {
        retVec[num] = 1;
      }

      rangeFlag = 0;
      index = 0;
      start = 0;

      if (strIndex > len) break;
    } else {
      if (c1 == '-') {
      
        /* Get previous number */
        chr[index] = END_OF_STR;
        start = (int) strtol(chr, &end, 10);
        if ((start < 1) || (start > n)) {
          printf_s("\nERROR: The list %s is not valid\n", str);
          exit(1);
        }

        rangeFlag = 1;
        index = 0;
      
      } else {
        if (index > im1) {
          printf_s("\nERROR: The list %s is not valid\n", str);
          exit(1);
        }
        chr[index] = c1;
        index++;
      }

    }
    strIndex++;
  }

  /* Get the maximum chr number */
  for (i=n-1; i>-1; i--) {
    if (retVec[i]) return(i);
  }

  return(0);

} /* END: getIntsFromStr */ 

/* Function to check a list of variable names in a sorted character vector */
int checkVarsInSortVec(str, sep, cvec, n, found, colNum)
char *str, sep, **cvec;
int n;
binary *found; /* Length should be n + 1 */
int *colNum; /* column number for 1 var in str */
{
  int i, str_n, tmp, nfound=0;
  char **str_vec, *end;
  binary cvecFlag, foundFlag, str1Flag;

  *colNum  = -1;

  /* Parse the str into a vector */
  str_vec = strIntoVec(str, sep, &str_n);

  if (str_n < 1) {
    printf_s("\nERROR with list: %s\n", str);
    exit(1);
  }

  cvecFlag  = (cvec != NULL) ? 1 : 0;
  foundFlag = (found != NULL) ? 1 : 0;
  str1Flag  = (str_n == 1) ? 1 : 0;

  /* Loop over each element */
  for (i=0; i<str_n; i++) {
    tmp = -1;
    if (cvecFlag) {
      tmp = findStrInVec(str_vec[i], cvec, n);
      /* Add 1 for column number */
      if (tmp > -1) tmp += 1;
      if (str1Flag) *colNum = tmp; 
    }
    if (tmp < 0) {
      /* Check if it is an integer */
      tmp = strtol(str_vec[i], &end, 10);
      if ((tmp <= 0) || (tmp > n)) tmp = -1;
      if (str1Flag) *colNum = tmp; 
    }
    if (tmp < 0) {
      printf_s("\nERROR: the variable %s was not found\n", str_vec[i]);
      exit(1);
    } else {
      if (foundFlag) found[tmp] = 1;
      nfound++;
    }
  }

  matrix_free((void **) str_vec, str_n);

  return(nfound);

} /* END: checkListCharSortVec */

/*****************************************************************************
***************** Reading files **********************************************
******************************************************************************/

/* Function for opening a data set in read mode */
FILE * open_file(datafile, mode)
char *datafile, *mode;
{
  FILE *fpt=0;

  fpt = fopen(datafile, mode);
  if (fpt == 0) {
    printf_s("\n Error opening file: %s \n", datafile);
    exit(0);
  }

  return(fpt);
}

/* Function for getting the number of columns in a delimited data file */
int getNcols(datafile, sep, bufferSize)
char *datafile, *sep;
long bufferSize;
{
  FILE *fpt;
  int i ;
  char line[bufferSize];

  fpt = open_file(datafile, "r");
  fgets(line, sizeof(line), fpt);
  fclose(fpt);
  i = count_char(line, *sep) + 1;

  return(i);
}

/* Function to determine the file delimiter */
char getDelimiter(datafile, bufferSize, gfpt)
char *datafile;
long bufferSize;
FILEFPT *gfpt;
{
  int i, len, n_tab=0, n_comma=0, n_space=0;
  char line[bufferSize], *pc, c1;

  my_fopen(datafile, gfpt);
  my_fgets(line, sizeof(line), gfpt);
  my_fclose(gfpt);

  len = (int) strlen(line);
  for (i=0, pc=line; i<len; i++, pc++) {
    c1 = *pc;
    if (c1 == '\t') {
      n_tab++;
    } else if (c1 == ' ') {
      n_space++;
    } else if (c1 == ',') {
      n_comma++;
    } 
  }

  if (n_tab) {
    return('\t');
  } else if (n_comma) {
    return(',');
  } else if (n_space) {
    return(' ');
  } else {
    return('\t');
  }

} /* getDelimiter */

/* Function to determine if a file has a header of variable names */
int getHeader(datafile, sep, ncols, bufferSize)
char *datafile, sep;
int ncols;
long bufferSize;
{
  FILE *fpt;
  int i, ret=0;
  char line[bufferSize], c1, *token, *end, cnum[2]; 

  fpt = open_file(datafile, "r");
  fgets(line, sizeof(line), fpt);
  fclose(fpt);

  cnum[1] = END_OF_STR;
  c1 = line[0];
  cnum[0] = c1;
  if ((c1 != '0') && (c1 != '.') && (c1 != '-') && (!strtol(cnum, &end, 10))) {
    ret = 1;
  } else {
    return(0);
  }
  token = line;
  for (i=1; i<ncols; i++) {
    token = next_token(token, bufferSize, sep);
    if (token == NULL) return(ret);
    c1 = token[0];
    cnum[0] = c1;
    if ((c1 != '0') && (c1 != '.') && (c1 != '-') && (!strtol(cnum, &end, 10))) {
      ret = 1;
    } else {
      return(0);
    }
  }

  return(ret);

} /* END: getHeader */

/* Function for getting the number of rows in a data file */
int getNrows(datafile, bufferSize)
char *datafile;
long bufferSize;
{
  FILE *fpt;
  int i = 0;
  char line[bufferSize];
  size_t size;

  size = sizeof(line);
  fpt = open_file(datafile, "r");
  while(fgets(line, size, fpt) != NULL) i++;
  fclose(fpt);
  return(i);
}

/* Function to read a column of ids */
char ** readIds(datafile, bufferSize, rmWhite, sort, retLen)
char *datafile;
long bufferSize;
int rmWhite, sort;
int *retLen;
{
  FILE *fpt;
  int i, j, nr, len, *tmpvec;
  char line[bufferSize], **vec, *id;
  size_t size;

  /* Get the number of ids */
  nr = getNrows(datafile, bufferSize);
  if (nr < 1) {
    printf_s("\nERROR: with file: %s \n", datafile);
    exit(1);
  }

  vec = (char **) malloc(nr*sizeof(char *)); 
  CHECK_MEM(vec);
  size = sizeof(line);

  i = 0;
  fpt = open_file(datafile, "r");
  while(fgets(line, size, fpt) != NULL) {
    if (rmWhite) {
      id  = rmWhiteSpace(line);
    } else {
      id = line;
    }
    len = strlen(id);
    vec[i] = (char *) malloc((len+1)*sizeof(char));
    CHECK_MEM(vec[i]);
    strcpy(vec[i], id);
    i++;
  }
  fclose(fpt);

  if (sort) {
    tmpvec = (int *) malloc(nr*sizeof(int));
    CHECK_MEM(tmpvec);
    for (j=0; j<nr; j++) tmpvec[j] = j;
    quicksort_str(0, nr-1, vec, tmpvec);
    free(tmpvec);
  }

  *retLen = nr;
  return(vec);

} /* END: readIds */

/* Function to get the column names in a file */
char ** getColNames(datafile, sep, bufferSize, sort, ncol)
char *datafile, sep;
long bufferSize;
int sort;
int *ncol;
{
  FILE *fpt;
  int nc, *tmpvec, j, *pi;
  char line[bufferSize], **ret; 

  fpt = open_file(datafile, "r");
  fgets(line, sizeof(line), fpt);
  fclose(fpt);

  ret = strIntoVec(line, sep, &nc);
  *ncol = nc;

  if (sort) {
    tmpvec = (int *) malloc(nc*sizeof(int));
    CHECK_MEM(tmpvec);
    for (j=0, pi=tmpvec; j<nc; j++, pi++) *pi = j;
    quicksort_str(0, nc-1, ret, tmpvec);
    free(tmpvec);
  }

  return(ret);
}

/*****************************************************************************
***************** Genotypes **************************************************
******************************************************************************/

/* Function to interchange 0 and 2 in a vector of genotypes */
void switch_genos(vec, n)
genoType *vec;  /* Coded as 0-1-2 and MISS_GENO */
int n;
{
  int i;
  genoType *p, geno;

  for (i=0, p=vec; i<n; i++, p++) {
    geno = *p;
    if (geno == 0) {
      *p = 2;
    } else if (geno == 2) {
      *p = 0;
    }
  }

} /* END: switch_genos */

/* Function to get the chr number from the alias */
int chrNumFromAlias(str, alias, nchr)
char *str, **alias;
int nchr;
{
  int i;

  /* Start loop index at 1 (no 0 chr) */
  if (alias) {
    for (i=1; i<nchr; i++) {
      if ((alias[i]) && (!strcmp(str, alias[i]))) return(i);
    }
  }

  return(ERROR_CHR);

} /* chrNumFromAlias */ 

/* Function to get the chromosome number of a string */
int getChrNum(str, nchr, flag, alias)
char *str, **alias;
int nchr, flag;
{
  int chr, len;
  char *end, c1, c2; 

  if ((flag) && (alias)) {
    chr = chrNumFromAlias(str, alias, nchr);
    if (chr != ERROR_CHR) return(chr);
  }

  chr = (int) strtol(str, &end, 10);
  if (chr) {
    if (chr < nchr) {
      return(chr);
    } else {
      return(ERROR_CHR);
    }
  } else {
    len = strlen(str);
    c1 = toupper(str[0]);
    if (len == 1) {
      if (c1 == 'X') {
        return(23);
      } else if (c1 == 'Y') {
        return(24);
      } else {
        return(ERROR_CHR);
      }
    } else if (len == 2) {
      c2 = toupper(str[1]);
      if ((c1 == 'X') && (c2 == 'Y')) {
        return(25);
      } else if ((c1 == 'M') && (c2 == 'T')) {
        return(26);
      } else {
        return(ERROR_CHR);
      }
    } else {
      return(ERROR_CHR);
    }
  }

} /* END: getChrNum */

/* Function to get the chromosome numbers from a string such as "1-7,X,21" */
int getChrsFromStr(str, n, retVec, alias, n_alias)
char *str;
int n;  /* Length of retVec (1 + max number) */
binary *retVec;
char **alias;
int n_alias; /* Length of alias */
{
  int i, strIndex, len, index, start, num, rangeFlag, im1, intLen=2;
  char chr[5], c1;

  /* Initialize */
  for (i=0; i<n; i++) retVec[i] = 0;
  len = (int) strlen(str);
  
  strIndex = 0;
  start = 0;
  rangeFlag = 0;
  index = 0;
  c1 = 'e';
  im1 = intLen - 1;

  while(1) {
    if (strIndex <= len) c1 = str[strIndex];

    if ((strIndex > len) || (c1 == ',')) {

      if (!index) {
        printf_s("\nERROR: 1. The list %s is not valid\n", str);
        exit(1);
      }

      /* Get number */
      chr[index] = END_OF_STR;
      num = getChrNum(chr, n_alias, 0, alias);
      if (num == ERROR_CHR) num = getChrNum(chr, n_alias, 1, alias);
      if ((num < 1) || (num > n)) {
        printf_s("\nERROR: 2. The list %s is not valid\n", str);
        exit(1);
      }

      if (rangeFlag) {
        if ((!start) || (num < start)) {
          printf_s("\nERROR: 3. The list %s is not valid\n", str);
          exit(1);
        }
        for (i=start; i<=num; i++) retVec[i] = 1;
      } else {
        retVec[num] = 1;
      }

      rangeFlag = 0;
      index = 0;
      start = 0;

      if (strIndex > len) break;
    } else {
      if (c1 == '-') {
      
        /* Get previous number */
        chr[index] = END_OF_STR;
        start = getChrNum(chr, n_alias, 0, alias);
        if (start == ERROR_CHR) start = getChrNum(chr, n_alias, 1, alias);
        if ((start < 1) || (start > n)) {
          printf_s("\nERROR: 4. The list %s is not valid\n", str);
          exit(1);
        }

        rangeFlag = 1;
        index = 0;
      
      } else {
        if (index > im1) {
          if (c1 != END_OF_STR) {
            printf_s("\nERROR: 5. The list %s is not valid\n", str);
            exit(1);
          }
        }
        chr[index] = c1;
        index++;
      }

    }
    strIndex++;
  }

  /* Get the maximum chr number */
  for (i=n-1; i>-1; i--) {
    if (retVec[i]) return(i);
  }

  return(0);

} /* END: getChrsFromStr */ 

/*******************************************************************/
/******** Vectors and matrices *************************************/
/*******************************************************************/

/* Function to compute the minimum of a double vector */
double min_dvec(vec, n)
double *vec;
int n;
{
  int i;
  double *p, min, temp;

  min = *vec;
  if (n > 1) {
    for (i=1, p=&vec[1]; i<n; i++, p++) {
      temp = *p;
      if (temp < min) min = temp;
    }
  }

  return(min);

} /* END: min_dvec */

/* Function to compute the minimum of an integer vector */
int min_ivec(vec, n)
int *vec, n;
{
  int i, *p, min, temp;

  min = *vec;
  if (n > 1) {
    for (i=1, p=&vec[1]; i<n; i++, p++) {
      temp = *p;
      if (temp < min) min = temp;
    }
  }

  return(min);

} /* END: min_ivec */

/* Function to compute the mean of a double vector */
double mean_dvec(vec, n)
double *vec;
int n;
{
  int i;
  double sum=0.0, *p;

  for (i=0, p=vec; i<n; i++, p++) sum += *p;
  sum = sum/n;

  return(sum);

} /* END: mean_dvec */

/* Function to compute xx` for a double column vector x */
void xxp(x, n, mat)
double *x;
int n;
double **mat;  /* Return matrix of dimension nxn */
{
  int i, j;
  double xi, temp;

  for (i=0; i<n; i++) {
    xi = x[i];
    mat[i][i] = xi*xi; 
    for (j=i+1; j<n; j++) {
      temp      = xi*x[j];
      mat[i][j] = temp;
      mat[j][i] = temp;  
    }
  }

} /* END: xxp */

/* Function to compute X`X with no missing values */
void Xp_X(X, nr, nc, mat)
double **X, **mat;
int nr, nc;
{
  int i, j, k;
  double sum, temp1;

  for (i=0; i<nc; i++) {
    sum = 0.0;
    /* Diagonal elements */
    for (k=0; k<nr; k++) {
      temp1 = X[k][i];
      sum += temp1*temp1; 
    }
    mat[i][i] = sum;

    /* Off diagonal */
    for (j=i+1; j<nc; j++) {
      sum = 0.0;
      for (k=0; k<nr; k++) sum += X[k][i]*X[k][j];
      mat[i][j] = sum;
      mat[j][i] = sum;  
    }
  }

} /* END: Xp_X */

/* Fucntion to compute X`y for matrix X and vector y */
void Xp_y(X, nr, nc, y, ret)
double **X, *y, *ret;
int nr, nc;
{
  int i, j;
  double sum, *p, *pret;

  for (i=0, pret=ret; i<nc; i++, pret++) {
    sum = 0.0;
    for (j=0, p=y; j<nr; j++, p++) sum += X[j][i]* *p;
    *pret = sum; 
  }

} /* END: Xp_y */

/* Function to compute XX` for a matrix X with missing values */
void X_Xp_miss(X, nr, nc, miss, mat)
double **X, miss;
int nr, nc;
double **mat;
{
  int i, j, k;
  double sum, *p, temp1, temp2, temp, *p2;

  for (i=0; i<nr; i++) {
    sum = 0.0;
    /* Diagonal elements */
    for (k=0, p=X[i]; k<nc; k++, p++) {
      temp1 = *p;
      if (fabs(temp1 - miss) > ALMOST_ZERO) sum += temp1*temp1; 
    }
    mat[i][i] = sum;

    /* Off diagonal */
    for (j=i+1; j<nr; j++) {
      sum = 0.0;
      for (k=0, p=X[i], p2=X[j]; k<nc; k++, p++, p2++) {
        temp1 = *p;
        temp2 = *p2;
        if ((fabs(temp1 - miss) > ALMOST_ZERO) && (fabs(temp2 - miss) > ALMOST_ZERO)) {
          temp = temp1*temp2;
          sum += temp;
        } 
      }
      mat[i][j] = sum;
      mat[j][i] = sum;  
    }
  }

} /* END: X_Xp_miss */

/* Function to compute [D(V)^(-1/2)]V[D(V)^(-1/2)] for a covariance matrix V */
void DV_V_DV(V, n, mat)
double **V;
int n;
double **mat; /* return matrix */
{
  int i, j;
  double D[n], *pd, Di, temp;

  /* compute D(V)^(-1/2) */
  for (i=0, pd=D; i<n; i++, pd++) *pd = 1.0/sqrt(V[i][i]);
  
  for (i=0, pd=D; i<n; i++, pd++) {
    Di = *pd;
    mat[i][i] = Di*Di*V[i][i];
    for (j=i+1; j<n; j++) {
      temp = Di*D[j]*V[i][j];
      mat[i][j] = temp;
      mat[j][i] = temp;
    }
  }

} /* END: DV_V_DV */

/* Multiply to matrices (matrix mult)  */
static void matrixMult(m1, m1_nr, m1_nc, m2, m2_nc, ret)
double **m1, **m2, **ret;
int m1_nr, m1_nc, m2_nc;
{
  int i, j, k;
  double sum;

  for (i=0; i<m1_nr; i++) {
    for (j=0; j<m2_nc; j++) {
      sum = 0.;
      for (k=0; k<m1_nc; k++) sum += m1[i][k]*m2[k][j]; 
      ret[i][j] = sum;
    }
  }

} /* END: matrixMult */

/* Factor a symmetric positive definite matrix */
static int symPosFactor(mat, n, ret, retdiag)
double **mat, **ret, *retdiag;
int n;
{
  int i, j, k;
  double sum, save, *ptr;
  
  /* Copy mat to ret */
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) ret[i][j] = mat[i][j];
  }

  for (i=0, ptr=retdiag; i<n; i++, ptr++) {
    for (j=i; j<n; j++) {
      sum = ret[i][j];
      for (k=i-1; k>-1; k--) sum -= ret[i][k]*ret[j][k];
      if (i == j) {
        if (sum < NUM_ZERO) return(ERROR_SINGULAR_MATRIX); 
        save = sqrt(sum);
        *ptr = save;
      } else {
        ret[j][i] = sum/save;
      }
    }
  }

  /* Zero out the diagonal and above */
  for (i=0; i<n; i++) {
    for (j=i; j<n; j++) ret[i][j] = 0.;
  }

  return(0);

} /* END: symPosFactor */

/* Invert a factor */
static void symPosFacInv(L, diag, n, ret)
double **L, *diag;
int n;
double **ret;
{
  int i, j, k;
  double sum;
  
  /* Copy L to ret */
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) ret[i][j] = L[i][j];
  }

  for (i=0; i<n; i++) {
    ret[i][i] = 1./diag[i];
    for (j=i+1; j<n; j++) {
      sum = 0.;
      for (k=i; k<=j-1; k++) sum -= ret[j][k]*ret[k][i];
   
      ret[j][i] = sum/diag[j];
    }
  }

} /* END: symPosFacInv */

/* Transpose of a square matrix */
static void matTranspose(mat, n, ret)
double **mat;
int n;
double **ret;
{
  int i, j;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) ret[j][i] = mat[i][j];   
  }

} /* END: rmatTranspose */

/* Inverse of symmetric positive definite matrix */
static int symPosMatInv(mat, n, ret)
double **mat; 
int n;
double **ret;
{
  double **L, **Linv, *diag;  
  int i;

  L = dblMatrix_alloc(n, n);
  diag = (double *) malloc(n*sizeof(double));
  CHECK_MEM(diag);

  /* Get cholesky lower triangle */
  i = symPosFactor(mat, n, L, diag);
  if (i) {
    free(diag);
    matrix_free((void **) L, n);
    return(i);
  }

  /* Get the lower inverse */
  Linv = dblMatrix_alloc(n, n);
  symPosFacInv(L, diag, n, Linv);
  free(diag);

  /* Get the transpose of Linv */
  matTranspose(Linv, n, L);

  /* Inverse is t(L^-1)(L^-1) */
  matrixMult(L, n, n, Linv, n, ret);

  matrix_free((void **) L, n);
  matrix_free((void **) Linv, n);

  return(0);

} /* END: symPosMatInv */

/* Function to compute the inverse of a covariance matrix */
int cov_inv(cov, n, inv)
double **cov;
int n;
double **inv; /* Returned inverse */
{
  double cc, a, b, d;
  int ret;

  switch (n) {
    case 0:
      printf_s("\nERROR: dimension of covariance matrix is 0\n", NULL);
      exit(1);
    case 1:
      a = cov[0][0];
      if (fabs(a) < ALMOST_ZERO) return(ERROR_SINGULAR_MATRIX);
      inv[0][0] = 1.0/a; 
      break;
    case 2:
      a  = cov[0][0];
      b  = cov[0][1];
      d  = cov[1][1];
      cc = a*d - b*b;
      if (fabs(cc) < ALMOST_ZERO) return(ERROR_SINGULAR_MATRIX);
      cc = 1.0/cc;
      inv[0][0] = d*cc;
      inv[0][1] = -b*cc;
      inv[1][0] = -b*cc;
      inv[1][1] = a*cc;
      break;
    default:
      ret = symPosMatInv(cov, n, inv);
      if (ret) return(ret);
      break; 
  } /* END: switch */

  return(0);

} /* END: cov_inv */

/* Function to invert a matrix */
void matrixInverse(mat, n, ret, error)
double **mat, **ret;
int n, *error;
{
  int i, j, r, *p, *ptri, k, hi;
  double max, temp, *hv, hr;

  /* Copy mat to ret */
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) ret[i][j] = mat[i][j];
  }

  p = (int *) malloc(n*sizeof(int));
  CHECK_MEM(p);
  hv = (double *) malloc(n*sizeof(double));
  CHECK_MEM(hv);
 
  for (j=0, ptri=p; j<n; j++, ptri++) *ptri = j; 
  for (j=0; j<n; j++) {

    /* pivot search */
    max = fabs(ret[j][j]);
    r = j;
    for (i=j+1; i<n; i++) {
      temp = fabs(ret[i][j]);
      if (temp > max) {
        max = temp;
        r = i;
      }
    }

    if (max < ALMOST_ZERO) {
      *error = ERROR_SINGULAR_MATRIX;
      return;
    }
 
    /* Row interchange */
    if (r > j) {
      for (k=0; k<n; k++) {
        hr = ret[j][k];
        ret[j][k] = ret[r][k];
        ret[r][k] = hr;
      }
      hi = p[j];
      p[j] = p[r];
      p[r] = hi; 
    }

    /* Transformation */
    hr = 1.0/ret[j][j];
    for (i=0; i<n; i++) ret[i][j] = hr*ret[i][j];
    ret[j][j] = hr;
    for (k=0; k<n; k++) {
      if (k != j) {
        for (i=0; i<n; i++) {
          if (i != j) {
            ret[i][k] = ret[i][k] - ret[i][j]*ret[j][k];
          }
        }
        ret[j][k] = -hr*ret[j][k];
      }
    }

  } /* END: for (j=0; j<n; j++) */

  /* Column interchange */
  for (i=0; i<n; i++) {
    for (k=0; k<n; k++) hv[p[k]] = ret[i][k];
    for (k=0; k<n; k++) ret[i][k] = hv[k];
  }

  free(p);
  *error = 0;

} /* END: matrixInverse */

/* Function to compute the inverse of a matrix */
int mat_inv(cov, n, inv)
double **cov;
int n;
double **inv; /* Returned inverse */
{
  double cc, a, b, c, d;
  int error;

  switch (n) {
    case 0:
      printf_s("\nERROR: dimension of covariance matrix is 0\n", NULL);
      exit(1);
    case 1:
      a = cov[0][0];
      if (fabs(a) < ALMOST_ZERO) return(ERROR_SINGULAR_MATRIX);
      inv[0][0] = 1.0/a; 
      break;
    case 2:
      a  = cov[0][0];
      b  = cov[0][1];
      c  = cov[1][0];
      d  = cov[1][1];
      cc = a*d - b*c;
      if (fabs(cc) < ALMOST_ZERO) return(ERROR_SINGULAR_MATRIX);
      cc = 1.0/cc;
      inv[0][0] = d*cc;
      inv[0][1] = -b*cc;
      inv[1][0] = -c*cc;
      inv[1][1] = a*cc;
      break;
    default:
      matrixInverse(cov, n, inv, &error);
      if (error) return(error);
      break; 
  } /* END: switch */

  return(0);

} /* END: mat_inv */

/* Function to compute beta = inv(X`X)X`y */
int beta_normal(X, nr, nc, y, beta, xpx)
double **X, *y, *beta;
int nr, nc;
double **xpx;
{
  double **mat, **inv, *vec;
  int ret, flag;
  
  /* Check for no covariates */
  if (X == NULL) {
    *beta = mean_dvec(y, nr);
    return(0);
  }

  inv = dblMatrix_alloc(nc, nc);
  vec = (double *) malloc(nc*sizeof(double));
  CHECK_MEM(vec);

  if (!xpx) {
    flag = 0;
    mat = dblMatrix_alloc(nc, nc);

    /* Compute X`X */
    Xp_X(X, nr, nc, mat);
  } else {
    mat = xpx;
    flag = 1;
  }

  /* Compute the inverse */
  ret = mat_inv(mat, nc, inv);
  if (!flag) matrix_free((void **) mat, nc);

  /* Compute X`y */
  Xp_y(X, nr, nc, y, vec);

  /* Compute beta */
  Xp_y(inv, nc, nc, vec, beta);

  free(vec);
  matrix_free((void **) inv, nc);

  return(ret);

} /* END: beta_normal */

/* Function to compute beta for logistic reg */
int beta_logReg(X, nr, nc, y, beta, xpx)
double **X, *y, *beta;
int nr, nc;
double **xpx;
{
  double **mat, **inv, *vec, temp, *beta0, *p, *p1, sum, *p2, *pvec, *vec2, sum2;
  int i, ii, j, k, ret=0, flag;  

  /* Check for no covariates */
  if (X == NULL) {
    temp = mean_dvec(y, nr);
    *beta = log(temp/(1.0 - temp)); 
    return(0);
  }

  mat = dblMatrix_alloc(nc, nc);
  inv = dblMatrix_alloc(nc, nc);
  beta0 = (double *) malloc(nc*sizeof(double));
  CHECK_MEM(beta0);
  vec = (double *) malloc(nr*sizeof(double));
  CHECK_MEM(vec);
  pvec = (double *) malloc(nr*sizeof(double));
  CHECK_MEM(pvec);
  vec2 = (double *) malloc(nc*sizeof(double));
  CHECK_MEM(vec2);

  if (!xpx) {
    flag = 0;
    xpx = dblMatrix_alloc(nc, nc);

    /* Compute X`X */
    Xp_X(X, nr, nc, xpx);
  } else {
    flag = 1;
  }

  /* Initial estimates [(X`X)^-1]X`Z, where Z = 2Y-1 */
  for (i=0, p1=vec, p2=y; i<nr; i++, p1++, p2++) *p1 = 2.0 * *p2 - 1.0;
  ret = beta_normal(X, nr, nc, vec, beta0, xpx);
  if (ret) return(ret);

  /* Loop until convergence */
  for (ii=0; ii<LOGREG_MAXIT; ii++) {

    /* Compute pi */
    for (i=0; i<nr; i++) {
      /* Compute X*beta */
      sum = 0.0;
      for (j=0, p=X[i], p1=beta0; j<nc; j++, p++, p1++) sum += *p * *p1;
          
      /* Compute 1/(1+ exp(-x)) */
      sum = 1.0/(1.0 + exp(-sum));

      /* Save pi */
      pvec[i] = sum;
      sum = sum*(1.0 - sum); 
      
      /* Check if sum is 0 or 1 */
      if (sum < LOGREG_ZERO) {
        sum = LOGREG_ZERO;
      } else if (sum > LOGREG_ONE) {
        sum = LOGREG_ONE;
      }
      vec[i] = sum;
    }

    for (i=0; i<nc; i++) {
      for (j=0; j<nc; j++) {
        sum  = 0.0;
        sum2 = 0.0;
        for (k=0; k<nr; k++) {
          temp = X[k][j];
          sum += temp*X[k][i]*vec[k];
 
          /* For t(X) * (y-pi) */
          if (!i) sum2 += temp*(y[k]-pvec[k]); 
        }
        mat[i][j] = sum;
        if (!i) vec2[j] = sum2;
      }
      mat[i][i] += LOGREG_FUZZ;
    }

    /* Invert mat */
    ret = mat_inv(mat, nc, inv);
    if (ret) break;

    /* Update beta, beta0 and check convergence */
    sum2 = 0.0;
    for (i=0; i<nc; i++) {
      sum = 0.0;
      for (j=0, p=inv[i], p2=vec2; j<nc; j++, p++, p2++) sum += *p * *p2;
      temp = beta0[i];
      beta[i] = temp + sum;

      sum2 += fabs(sum);
      beta0[i] = beta[i];
    }

    if (sum2 < LOGREG_EPS) break;
  }   
  if (ii >= LOGREG_MAXIT) ret = ERROR_MAXITER;

  free(pvec);
  free(vec);
  free(beta0);
  matrix_free((void **) mat, nc);
  matrix_free((void **) inv, nc);
  if (!flag) matrix_free((void **) xpx, nc);

  return(ret);

} /* END: beta_logReg */

/* Function to compute E(Y | X) for normal y */
int EY_X(X, nr, nc, y, ccFlag, rety)
double **X, *y, *rety;
int nr, nc;
binary ccFlag;  /* 0=continuous y, 1=binary */
{
  double *beta, *pb, *pret, sum, *px, b1, temp;
  int i, j, Xflag;

  beta = (double *) malloc(nc*sizeof(double));
  CHECK_MEM(beta);

  /* Compute the parameter estimates */
  if (ccFlag) {
    i = beta_logReg(X, nr, nc, y, beta, NULL);
  } else {
    i = beta_normal(X, nr, nc, y, beta, NULL);
  }
  if (i) {
    free(beta);
    return(i);
  }

  if (X != NULL) {
    Xflag = 1;
  } else {
    Xflag = 0;
    if (ccFlag) {
      temp = exp(*beta);
      b1 = temp/(1.0 + temp);
    } else {
      b1 = *beta;
    }
  }

  /* Compute the linear predictor for each subject */
  for (i=0, pret=rety; i<nr; i++, pret++) {
    if (Xflag) {
      sum = 0.0;
      for (j=0, pb=beta, px=X[i]; j<nc; j++, pb++, px++) sum += *px * *pb;
      if (ccFlag) {
        temp = exp(sum);
        sum  = temp/(1.0 + temp);
      }  
      *pret = sum;
    } else {
      *pret = b1;
    }
  } 

  free(beta);

  return(0);

} /* END: EY_X */

/* Function to normalize a double vector (subtract mean and divide by sd) */
void norm_dvec(vec, n, var_divide, ret)
double *vec, *ret;
int n;
int var_divide; /* Either n or n-1 to divide the variance */
{
  int i;
  double *p1, *p2, sum=0.0, sum2=0.0, temp, mean, sd;

  for (i=0, p1=vec; i<n; i++, p1++) {
    temp  = *p1;
    sum  += temp;
    sum2 += temp*temp;
  }
  mean = sum/n;
  sd   = sqrt((sum2 - n*mean*mean)/var_divide);
  sd   = 1.0/sd;

  for (i=0, p1=vec, p2=ret; i<n; i++, p1++, p2++) *p2 = (*p1 - mean)*sd;

} /* END: norm_dvec */

/* Function to copy a square matrix to another square matrix excluding certain rows/columns */
void copySqMat_exclude(mat, exclude, n, mat1, n1)
double **mat; /* Matrix to get update with rows, cols from mat1 */
binary *exclude; /* 0 or 1 to keep/exclude rows/col of mat from being copied */
int n; /* number of rows/cols in mat and length of exclude */
double **mat1; /* Elements in this matrix will be put in mat */
int n1; /* Number of rows/cols of mat1 */
{
  int i, j, nexclude=0, index;
  binary *pb, *pb2;
  double *p1, *p2;  

  /* Get the number of rows cols to exclude */
  for (i=0, pb=exclude; i<n; i++, pb++) {
    if (*pb) nexclude++;
  }

  if (!nexclude) return;

  index = -1;
  for (i=0, pb=exclude; i<n; i++, pb++) {
    if (!*pb) {
      index++;
      p2 = mat1[index];
      for (j=0, p1=mat[i], pb2=exclude; j<n; j++, p1++, pb2++) {
        if (!*pb2) *p1 = *p2++;
      }
    }
  }

} /* END: copySqMat_exclude */

/*******************************************************************/
/********************* Sorting *************************************/
/*******************************************************************/

/* For character strings 
#include <string.h>
i = strcmp( s1, s2 );

Where:

const char *s1, *s2;
    are the strings to be compared. 
int i;
    gives the results of the comparison. 
    "i" is zero if the strings are identical. 
    "i" is positive if string "s1" is greater than string "s2", and is negative if string "s2" is greater than string "s1". 
    Comparisons of "greater than" and "less than" are made according to the ASCII collating sequence. 
*/

/* Sorting function, x is changed, sort from small to large*/
void quicksort_str(start, stop, x, cvec)
int start;  /* Starting position, usually 0 */
int stop;   /* End position */
char **x;   /* Vector of strings */
int *cvec;  /* Initialize to 0, 1,... n-1 */
{
 int i, j, k;
 char *temp, *median;
 int tempd;

  while (start < stop) {
    /*
    ** first-- if the list is short, do an ordinary insertion sort
    */
    if ((stop-start)<11) {
	for (i=start+1; i<=stop; i++) {
	    temp  = x[i];
	    tempd = cvec[i];
	    j = i-1;

           /*while (j>=start && (x[j]>temp)) {*/
	    while (j>=start && (strcmp(x[j], temp) > 0)) {
		x[j+1] = x[j];
		cvec[j+1] = cvec[j];
		j--;
		}
	    x[j+1] = temp;
	    cvec[j+1]  = tempd;
	    }
	return;
	}

    /*
    ** list is longer -- split it into two
    **  I use the median of 3 values as the split point
    */
    i=start;
    j=stop;
    k = (start + stop)/2;

    median = x[k];

    /*
    if (x[i] >= x[k]) {      
      if (x[j] > x[k]) {   
        if (x[i] > x[j])  median = x[j];
	 else median= x[i];
      }
    } else {
      if (x[j] < x[k]) {
	 if (x[i] > x[j]) median = x[i];
	 else median = x[j];
      }
    }
    */

    if (strcmp(x[i], x[k]) >= 0) {      
      if (strcmp(x[j], x[k]) > 0) {   
        if (strcmp(x[i], x[j]) > 0)  median = x[j];
	 else median= x[i];
      }
    } else {
      if (strcmp(x[j], x[k]) < 0) {
	 if (strcmp(x[i], x[j]) > 0) median = x[i];
	 else median = x[j];
      }
    }

    /* 
    **  Now actually do the partitioning 
    **   Because we must have at least one element >= median, "i"
    **   will never run over the end of the array.  Similar logic
    **   applies to j.
    ** A note on the use of "<" rather than "<=".  If a list has lots
    **   of identical elements, e.g. 80/100 are "3.5", then we will
    **   often go to the swap step with x[i]=x[j]=median.  But we will
    **   get the pointers i and j to meet approximately in the middle of
    **   the list, and that is THE important condition for speed in a
    **   quicksort.
    **   
    */
    while (i<j) {
	/*
	** top pointer down till it points at something too large
	*/
	/*while (x[i] < median) i++;*/
       while (strcmp(x[i], median) < 0) i++;

	/*
	** bottom pointer up until it points at something too small
	*/
	/*while(x[j] > median) j--;*/
       while(strcmp(x[j], median) > 0) j--;

	if (i<j) {
	    /*if (x[i] > x[j]) {*/
           if (strcmp(x[i], x[j]) > 0) {
		temp = x[i];
		x[i] = x[j];
		x[j] = temp;
		tempd= cvec[i];   cvec[i] =cvec[j];  cvec[j] =tempd;
		}
	    i++; j--;
	    }
	}

    /*
    ** The while() step helps if there are lots of ties.  It will break
    **  the list into 3 parts: < median, ==median, >=median, of which only
    **  the top and bottom ones need further attention.
    ** The ">=" is needed because i may be  == to j
    */
    /*
    while (x[i] >= median && i>start) i--;
    while (x[j] <= median && j<stop ) j++;
    */

    while ( (strcmp(x[i], median) >= 0) && (i>start) ) i--;
    while ( (strcmp(x[j], median) <= 0) && (j<stop) ) j++;

    /*
    ** list has been split, now do a recursive call
    **   always recur on the shorter list, as this keeps the total
    **       depth of nested calls to less than log_base2(n).
    */
    if ((i-start) < (stop-j)) { /* top list is shorter */
	if ((i-start)>0) quicksort_str(start,i, x, cvec);
	start =j; 
	}

    else {    /* bottom list is shorter */
	if ((stop -j)>0) quicksort_str(j,stop, x, cvec);
	stop=i; 
	}
     }

} /* END: quicksort_str */

/* Function for a binary search of an integer in a sorted vector of integers */
int binSearch_int(vec, n , find)
int *vec, n, find;
{
  int i0, i1, mid, veci, nm1, max, count, mid0;

  veci = vec[0];
  if (find == veci) {
    return(0);
  } else if (find < veci) {
    return(-1);
  }
  nm1  = n - 1;
  veci = vec[nm1];
  if (find == veci) {
    return(nm1);
  } else if (find > veci) {
    return(-1);
  }

  i0    = 0;
  i1    = nm1;
  max   = i1 - i0 + 1;
  count = 0;
  mid0  = -1;
  while (1) {
    if (i1 - i0 < 2) {
      if (find == vec[i0]) {
        return(i0);
      } else if (find == vec[i1]) {
        return(i1);
      } else {
        return(-1); 
      }
    }

    mid = (i0 + i1)/2;
    if (mid == mid0) mid = ceil(0.5*(i0 + i1));
   
    veci = vec[mid];
    if (find > veci) {
      i0 = mid;
    } else if (find < veci) {
      i1 = mid; 
    } else if (find == veci) {
      return(mid);
    } 
 
    if (mid != mid0) {
      mid0 = mid;
    } else {
      return(-1);
    }
    if (count > max) return(-1);
    count++;
  }

  return(-1);  

} /* END: binSearch_int */


